from __future__ import annotations

import copy
import warnings
from numbers import Real
from typing import Union, Optional, Iterable, Tuple, Callable, List, Dict, Any

from matplotlib import pyplot as plt
from pydantic import BaseModel, field_validator, model_validator

from .hardware import (
    create_hollow_cone_beam,
    create_cone_of_acceptance,
    ID,
    OD,
    THETA,
    pencil_beams,
    total_acceptor,
)
from .utils import sample_spectrum
from .contrib.bio import calculate_mus
from .import_utils import np, iterable, NDArray

# Get some default properties
mu_s, mu_a, wl = calculate_mus()


class Medium(BaseModel):
    n: float = 1
    mu_s: Union[float, np.ndarray] = 0
    mu_a: Union[float, np.ndarray] = 0
    wavelengths: np.ndarray = wl
    ref_wavelength: float = 650
    g: float = 1
    desc: str = "default"
    display_color: Optional[str] = None

    class Config:
        arbitrary_types_allowed = True

    @field_validator("mu_s", "mu_a", "wavelengths", mode="before")
    @classmethod
    def convert_to_array(cls, v: Union[float, Iterable[float]]) -> np.ndarray:
        assert np.all(v >= 0)
        return np.array(v)

    @field_validator("g", mode="before")
    @classmethod
    def g_vals(cls, g: float) -> float:
        assert -1 <= g <= 1
        return g

    @model_validator(mode="after")
    def force_non_scatterers_g(self) -> Medium:
        if np.any(self.mu_s == 0 and self.g != 1):
            warnings.warn(
                "g is automatically set to 1 where mu_s is 0. "
                "Set a non-zero scattering coefficient if a non-unity g value is necessary."
            )
            self.g = np.where(self.mu_s == 0, 1, self.g)
        if iterable(self.g) and len(self.g) == 1:
            self.g = self.g[0]
        return self

    def __repr__(self):
        return self.desc.capitalize() + " Optical Medium Object"

    def __str__(self):
        if self.desc == "default":
            return f"Optical Medium: n={self.n}, mu_s={self.mu_s}, mu_a={self.mu_a}, g={self.g}"
        return self.desc.capitalize()

    def set(self, **kwargs: Any) -> None:
        """
        Sets multiple attributes of the object in bulk.

        This method is useful for updating the medium's attributes dynamically, particularly when modifying properties
        during iterative simulations while generating a lookup table. It assigns the provided keyword arguments to
        existing attributes of the object.

        :param kwargs: A dictionary of attribute names and their corresponding new values.
        :type kwargs: dict
        :return: None
        :raises AttributeError: If a provided attribute name does not exist.
        """
        for k, v in kwargs.items():
            setattr(self, k, v)

    def _wave_index(self, wavelength: Union[float, Iterable[float]]):
        if iterable(wavelength) and iterable(self.wavelengths):
            return [np.where(self.wavelengths == wl)[0][0] for wl in wavelength]
        elif iterable(self.wavelengths):
            return np.where(self.wavelengths == wavelength)[0][0]
        else:
            return 0

    def mu_s_at(self, wavelengths: float):
        if iterable(self.mu_s):
            return np.interp(wavelengths, self.wavelengths, self.mu_s)
        return self.mu_s

    def mu_a_at(self, wavelengths: float):
        if iterable(self.mu_a):
            return np.interp(wavelengths, self.wavelengths, self.mu_a)
        return self.mu_a

    def mu_t_at(self, wavelengths: float):
        if iterable(self.mu_t):
            return np.interp(wavelengths, self.wavelengths, self.mu_t)
        return self.mu_t

    def albedo_at(self, wavelengths: float):
        if iterable(self.albedo):
            return np.interp(wavelengths, self.wavelengths, self.albedo)
        return self.albedo

    @property
    def mu_t(self):
        return self.mu_s + self.mu_a

    @property
    def albedo(self):
        if self.mu_t == 0:
            return 0
        else:
            return self.mu_a / self.mu_t


class Illumination:
    def __init__(
        self,
        pattern: Callable = create_hollow_cone_beam((ID, OD), THETA),
        spectrum: Optional[Iterable[Real]] = None,
        desc: Optional[str] = None
    ) -> None:
        self.pattern = pattern
        self.spectrum = spectrum
        self.description = desc

    def photon(self, batch_size: int = 50000, **kwargs: Any) -> Photon:
        location, direction = self.pattern(batch_size)
        wavelength = (
            sample_spectrum(self.spectrum) if self.spectrum is not None else None
        )
        return Photon(
            wavelength,
            batch_size=batch_size,
            location_coordinates=location,
            directional_cosines=direction,
            **kwargs
        )

    def __repr__(self) -> str:
        return f"IlluminationObject(pattern={self.pattern}, spectrum={self.spectrum})"

    @property
    def desc(self) -> str:
        if self.description is None:
            return self.__repr__()
        return self.description


class Detector:
    def __init__(
        self,
        acceptor: Callable = create_cone_of_acceptance(ID),
        desc: Optional[str] = None,
    ) -> None:
        self.acceptor = acceptor
        self.n_total = 0
        self.n_detected = 0
        self.description = desc

    def detect(
        self,
        location: Iterable[float],
        direction: Union[float, Iterable[float]],
        weights: Optional[Union[float, Iterable[float]]] = None,
    ) -> None:
        weights = weights if weights is not None else 1
        self.n_total += np.nansum(weights)
        x, y = location[:, :2].T
        mu_z = direction[:, -1]
        accepted_mask = self.acceptor(x, y, mu_z=mu_z)
        self.n_detected += np.nansum(weights[accepted_mask])

    def __call__(
        self, photon: Photon, mask: Optional[NDArray[np.bool_]] = True
    ) -> None:
        if not isinstance(photon, Photon):
            raise ValueError(
                "Detector object can only be called directly with a photon. "
                "Use detector.detect() for non-photon test cases."
            )
        self.detect(
            photon.exit_location[mask],
            photon.exit_direction[mask],
            photon.exit_weights[mask],
        )

    def reset(self) -> None:
        self.n_total = 0
        self.n_detected = 0

    def __repr__(self) -> str:
        return f"Detector_object(acceptor=f{self.acceptor})"

    @property
    def desc(self) -> str:
        if self.description is None:
            return self.__repr__()
        return self.description


class System:
    def __init__(
        self,
        *args,
        surrounding_n: float = 1,
        illuminator: Optional[Illumination] = Illumination(pencil_beams),
        detector: Tuple[Detector, float] = (Detector(total_acceptor), 0)
    ) -> None:
        """
        Create a system of optical mediums and its surroundings that hold the optical properties and can determine the
        medium of a location as well as interfaces crossing given two locations. The blocks are constructed top down in
        the order they are received from the top down (positive z is downward) starting at 0 and surrounded by infinite
        surroundings.

        ### Process
        1. Create the surroundings using the input or default n
        2. Stack the surroundings from z=negative infinity to z=0
        3. Iterate through all *args, and create the following:
            - Dict of the system stack with Medium object keys and len=2 list of boundaries for respective
              object including surroundings
            - List of Medium object layers in order of stacking including surroundings
            - Ndarray of the boundary (i.e. interfaces) z location between Medium objects, excluding +- infinite
        4. Add surroundings to the bottom from z=system thickness to z= positive infinity if the last layer was not
        semi-infinite

        ### Paramaters
        :param *args: A variable number of ordered pairs of Medium objects and their respective thickness (in
        that order). The number of args input must be even.
        """
        self.illuminator = illuminator

        self.detector = detector[0]
        self.detector_location = detector[1]
        self._boundaries = [float("-inf"), 0, float("inf")]

        if not len(args) % 2 == 0:
            raise ValueError(
                "Arguments must be in groups of 2: medium object and thickness."
            )

        # Add surroundings from -inf to 0 to inf
        self.surroundings = Medium(n=surrounding_n, mu_s=0, mu_a=0, desc="surroundings")
        self.layer = [
            self.surroundings,
            self.surroundings,
        ]  # List of layers in order of addition

        # Iterate through args to stack layers
        for i in range(0, len(args), 2):
            self.add(args[i], args[i + 1])

    @property
    def boundaries(self) -> NDArray[float]:
        return np.asarray(self._boundaries)

    @boundaries.setter
    def boundaries(self, value: List):
        self._boundaries = value

    def __repr__(self) -> str:
        return (
            "|".join(
                [f" <-- {bound[0]} --> {layer}" for bound, layer in self.stack.items()]
            )
            + f" <-- {self.boundaries[-1]}|"
        )

    def __str__(self) -> str:
        stack = "\n"
        space = 0
        for bound, layer in self.stack.items():
            txt = layer.__str__()
            space = len(txt) if len(txt) > space else space
            txt = f"{bound[0]:.4g}"
            space = len(txt) if len(txt) > space else space
        txt = f"{bound[1]:.4g}"
        space = len(txt) if len(txt) > space else space
        space += 8

        for bound, layer in self.stack.items():
            txt = f" {bound[0]:.4g} "
            lfill = "-" * ((space - len(txt) - 2) // 2)
            rfill = "-" * (space - len(txt) - len(lfill) - 2)
            boundary = f"|{lfill}{txt}{rfill}|\n"

            txt = f" {layer} "
            lfill = " " * int(np.floor((space - len(txt) - 6) / 2))
            rfill = " " * (space - len(txt) - len(lfill) - 6)
            layer = f"|->{lfill}{txt}{rfill}<-|\n"
            stack += boundary + layer

        txt = f" {bound[1]:.4g} "
        lfill = "-" * ((space - len(txt) - 2) // 2)
        rfill = "-" * (space - len(txt) - len(lfill) - 2)
        boundary = f"|{lfill}{txt}{rfill}|\n"
        stack += boundary
        border = " " + "_" * (space - 2) + " "
        return border + stack + border

    @property
    def stack(self) -> Dict[Tuple[float, float], Medium]:
        """
        Provides an up-to-date dictionary representation of layer boundaries.

        This property dynamically constructs a dictionary mapping layer boundaries to their corresponding
        layers. It is useful for querying interfaces efficiently in float-time.

        :return: A dictionary where keys are tuples representing layer boundaries (lower_bound, upper_bound),
                 and values are the corresponding layers.
        :rtype: dict[tuple[float, float], Layer]
        """
        return {
            (self.boundaries[i], self.boundaries[i + 1]): layer
            for i, layer in enumerate(self.layer)
        }

    def add(self, layer: Medium, depth: float) -> None:
        """
        Adds a new layer of a given depth to the system.

        This method inserts a new `Medium` layer at the specified depth. If the preceding layer was
        semi-infinite, an error is raised. The method automatically replaces the surrounding filler layer
        before adding the new layer and ensures that the system remains properly bounded, extending back
        to infinity if necessary.

        :param layer: The `Medium` object representing the new layer.
        :type layer: Medium
        :param depth: The thickness of the new layer.
        :type depth: float
        :raises ValueError: If the preceding layer is semi-infinite.
        :return: None
        """

        if self.layer[-1] is not self.surroundings and self.boundaries[-1] == float(
            "inf"
        ):
            raise OverflowError("Cannot add to semi-infinite system.")
        depth = float(depth)

        # Get the bound of the last layer before surroundings
        replacement_idx = -2 if self.boundaries[-1] == float("inf") else -1
        start_interface = self.boundaries[replacement_idx]
        old_boundaries = self.boundaries[:replacement_idx]
        end_interface = start_interface + depth
        if replacement_idx == -2:
            self.layer[-1] = layer
        else:
            self.layer.append(layer)
        self.boundaries = np.append(old_boundaries, [start_interface, end_interface])

        # Add surroundings if not semi-infinte
        if end_interface < float("inf"):
            self.add(self.surroundings, float("inf"))

    def beam(self, batch_size: int = 50000, **kwargs: Any) -> Photon:
        photon = self.illuminator.photon(batch_size=batch_size, system=self)
        for key, val in kwargs.items():
            setattr(photon, key, val)
        return photon

    def in_medium(
        self, location: Iterable[float]
    ) -> NDArray[Union[Medium, Tuple[Medium, Medium]]]:
        """
        Return the medium(s) that are at the queried coordinate. If the coordinate is an interfaces location, the
        mediums that makeup the interfaces are returned as a tuple, this includes boundary interfaces being returned
        with False on the surroundings side.

        ### Process
        1. Get the z-coordinate from the input
        2. Check it against boundaries of each medium of the system until it is found that either:
            - It is between any of the boundaries, it is in the medium within those boundaries
            - It is at a boundary, it is "in" the two mediums that make up that interfaces
        3. Break and return the medium(s) of the queried point.

        ### Parameters:
        :param location: (tuple, list, ndarray, float, or int) The coordinates or z-coordinate to query.

        ### Returns
        :return in_: (medium or tuple of mediums): The medium of the queried z-coordinate or a tuple of interfaces if
                     the coordinate is at an interfaces
        """
        location = np.asarray(location)
        z = location if np.ndim(location) == 1 else np.asarray(location)[:, 2]
        in_medium = np.empty_like(z, dtype=object)
        in_medium = np.where(z == float("inf"), self.layer[-1], in_medium)
        in_medium = np.where(z == float("-inf"), self.layer[0], in_medium)

        for bound, medium in self.stack.items():
            # If between boundaries, get that medium
            mask_inside = np.logical_and(bound[0] < z, z < bound[1])
            mask_boundary = (bound[0] == z) & (~np.isinf(z))

            in_medium[mask_inside] = medium

            # If at a boundary, get the mediums on each side
            if np.any(mask_boundary):
                z_neg_move = np.nextafter(z[mask_boundary], float("-inf"))
                z_pos_move = np.nextafter(z[mask_boundary], float("inf"))
                output1 = self.in_medium(z_neg_move)
                output2 = self.in_medium(z_pos_move)
                for i, idx in enumerate(np.where(mask_boundary)[0]):
                    in_medium[idx] = (output1[i], output2[i])

            # If all have been filled in, break
            if not np.any(np.equal(in_medium, None)):
                break
        return in_medium

    def interface_crossed(
        self, location0: Iterable[float], location1: Iterable[float]
    ) -> Union[Tuple[Union[Medium, ()], Union[float, None]]]:
        """
        Determines the first interfaces crossed when moving between two locations, considering only the z-coordinates.

        This method checks if any interfaces boundaries lie between the given z-coordinates. If an interfaces is crossed,
        the method calculates its z-location and identifies the two mediums forming the interfaces.

        ### Process:
        1. Identify boundaries that fall between the start and end z-coordinates.
        2. Compute the distance from the start z-coordinate to each boundary.
        3. Apply a mask to filter boundaries that are actually crossed.
        4. Select the closest crossed boundary as the interfaces plane.
        5. Determine the two mediums making up the interfaces by slightly shifting the plane's z-coordinate:
           - Backwards (toward the start) to find the first medium.
           - Forwards (away from the start) to find the second medium.

        ### Parameters:
        - :param location1: Starting location of the query.
        - :param location2: Ending location of the query.

        ### Returns:
        - :return interfaces: (tuple): The two media forming the crossed interfaces, or an empty list `[]` if no
                              interfaces is crossed.
        - :return plane: (float or bool): The z-coordinate of the crossed interfaces if one is found, otherwise `False`.
        """

        # Get z coords
        z0 = (
            np.asarray(location0)
            if np.shape(location0)[0] == 1
            else np.asarray(location0)[:, 2]
        )
        z1 = (
            np.asarray(location1)
            if np.shape(location1)[0] == 1
            else np.asarray(location1)[:, 2]
        )

        # Move off of boundaries (uncrossed!)
        is_boundary1 = np.isin(z0, self.boundaries)
        is_boundary2 = np.isin(z1, self.boundaries)

        z0 = np.where(is_boundary1, np.nextafter(z0, z1), z0)
        z1 = np.where(is_boundary2, np.nextafter(z1, z0), z1)

        # Sort for easier logic
        z_sorted = np.sort(np.stack((z0, z1), axis=-1), axis=-1)

        # Check if any boundaries fall between the zs and put into nd boolean array to use as a mask
        boundaries = np.asarray(self.boundaries, dtype=np.float64)
        crossed_mask = (z_sorted[..., 0, np.newaxis] < boundaries) & (
            boundaries < z_sorted[..., 1, np.newaxis]
        )

        # Determine closest crossed boundary (if any)
        dist_from_start = np.abs(boundaries - z0[..., None])
        dist_from_start[~crossed_mask] = np.inf  # Ignore non-crossed boundaries

        closest_idx = np.argmin(dist_from_start, axis=-1)

        plane = np.where(
            np.any(crossed_mask, axis=-1), boundaries[closest_idx], None
        ).astype(np.float64)

        # Determine mediums at the interface
        has_interface = np.any(crossed_mask, axis=-1)
        if np.any(has_interface):
            interface0 = np.where(
                has_interface, self.in_medium(np.nextafter(plane, z0)), None
            )
            interface1 = np.where(
                has_interface, self.in_medium(np.nextafter(plane, z1)), None
            )

            # Combine mediums into tuples or return empty tuples where no interface was crossed
            interfaces = np.array(
                [
                    tuple(pair) if valid else ()
                    for valid, pair in zip(has_interface, zip(interface0, interface1))
                ],
                dtype=object,
            )
            return interfaces, plane
        else:
            return None, None

    def represent_on_axis(self, ax: plt.axes.Axes = None) -> None:
        if ax is None:
            ax = plt.gca()
            lim = (
                [-0.1 * self.boundaries[-1], 1.1 * self.boundaries[-1]]
                if self.boundaries[-1] != float("inf")
                else [-0.1 * ax.get_ylim()[0], 1.1 * ax.get_ylim()[1]]
            )
            if ax.name == "3d":
                ax.set(zlim=lim)
        else:
            alpha = 0
            for bound, medium in self.stack.items():
                depth = np.diff(bound)
                y_edge = bound[0] - 0.1 * depth
                x_edge = ax.get_xlim()[0] * 0.95
                ax.text(x_edge, y_edge, medium.desc, fontsize=12)
                line_x = 100 * np.asarray(ax.get_xlim())
                y1 = bound[0] if not np.isinf(bound[0]) else ax.get_ylim()[0]
                y2 = bound[1] if not np.isinf(bound[1]) else ax.get_ylim()[1]
                alpha += 0.2
                ax.fill_between(
                    line_x,
                    y1,
                    y2,
                    color=(
                        "gray" if medium.display_color is None else medium.display_color
                    ),
                    alpha=alpha if medium.display_color is None else 1,
                )


class IndexableProperty(np.ndarray):
    def __new__(
        cls, arr: Iterable, normalize: bool = False, dtype: Optional[np.dtype] = None
    ) -> IndexableProperty:
        obj = np.asarray(arr, dtype=dtype).view(cls)
        obj.normalize = normalize
        obj /= np.linalg.norm(obj, axis=-1)[..., np.newaxis] if normalize else 1
        return obj

    def __array_finalize__(self, obj: Optional[NDArray]):
        if obj is None:
            return None
        self._normalize = getattr(obj, "_normalize", False)

    @property
    def normalize(self):
        return self._normalize

    @normalize.setter
    def normalize(self, value: bool):
        self._normalize = value
        if self._normalize:
            self /= np.linalg.norm(self, axis=-1)[..., np.newaxis]

    def __setitem__(self, index: int, value: float):
        super().__setitem__(index, value)
        if self.normalize:
            self /= np.linalg.norm(self, axis=-1)[..., np.newaxis]

    def __getitem__(self, index: int) -> IndexableProperty[float]:
        item = super().__getitem__(index)
        if isinstance(item, IndexableProperty):
            item.normalize = False
        return item


# TODO: Make this work with 1 photon the same it does for batches
class Photon:
    class Photon:
        """
        Represents a photon or a batch of photons with properties and behaviors that simulate photon interactions
        with a medium, including movement, absorption, scattering, and more.

        This class simulates the behavior of photons in an optical system, including their movement through
        different mediums, interactions with interfaces, and absorption or scattering events. The photons are
        modeled as a batch, and each photon has attributes such as wavelength, directional cosines, location,
        and weights, which can be modified and tracked throughout the simulation.

        Attributes:
        - n (int): The number of photons in the batch.
        - wavelength (Union[float, Iterable[float]]): The wavelength(s) of the photon(s).
        - system (Optional[System]): The system that contains the photon batch.
        - directional_cosines (IndexableProperty[float]): The directional cosines for each photon in the batch.
        - location_coordinates (np.ndarray): The coordinates of the location of each photon in the batch.
        - weights (np.ndarray): The weights of each photon in the batch.
        - russian_roulette_constant (float): The constant used for photon survival in the Russian roulette process.
        - recurse (bool): Whether recursion is allowed in the simulation.
        - recursion_depth (int): The current depth of recursion in the simulation.
        - recursion_limit (float): The limit for recursion depth in the simulation.
        - throw_recursion_error (bool): Whether to throw an error when recursion limit is exceeded (alt. warn and break).
        - keep_secondary_photons (bool): Whether to retain secondary (recursed) photons in the simulation.
        - tir_limit (float): The limit for total internal reflection (TIR).
        - A (float): The amount of the batch that has been absorbed.
        - R (float): The amount of the batch that has been reflected (out of the system).
        - T (float): The amount of the batch that has been transmitted (out of the system).
        - exit_location (np.ndarray[float]): The coordinates of the photon at the end of the simulation.
        - exit_direciton (np.ndarray[float]): The direction of the photon at the end of the simulation.
        - exit_weights (np.ndarray[float]): The weights of the photon at the end of the simulation.
        - location_history (np.ndarray[float]): The location history of the photon at all steps in the batch simulation.
        - weights_history (np.ndarray[float]): The weights history of the photon at all steps in the batch simulation.
        - cache_register (np.ndarray[np.bool_]): A cache register to track whether medium needs to be re-queried returned from the cache.
        - at_interface (np.ndarray[np.bool_]): A boolean array indicating whether the photon is at an interface currently.


        Methods:
        - __init__: Constructor for the photon that handles implicit batching necessities.
        - simulate: Runs the photon simulation until all photons are terminated.
        - absorb: Decrements each photon's weights based on current medium's albedo.
        - move: Moves the photons one step, handling interface crossing internally.
        - reflect_refract: Deterministically updates the direciton of photons based on interface properties.
        - scatter: Randomly updates the direction of each photon based on the scattering properties of the medium.
        - copy: Creates a deep copy of the photon object, with optional attribute overwrites.
        - russian_roulette: Simulates photon survival using the Russian roulette method.
        - __repr__: Provides a string representation of the photon object for debugging.
        - plot_path: A plotting function to plot the photon histories in 3d or projected onto an input plane.
        """

    def __init__(
        self,
        wavelength: Union[float, Iterable[float]],
        batch_size: int = 0,
        system: Optional[System] = None,
        directional_cosines: Iterable[float] = (0, 0, 1),
        location_coordinates: Iterable[float] = (0, 0, 0),
        weights: Union[float, Iterable[float]] = 1,
        russian_roulette_constant: float = 20,
        recurse: bool = True,
        recursion_depth: Optional[int] = 0,
        recursion_limit: Optional[float] = 100,
        throw_recursion_error: bool = True,
        keep_secondary_photons: bool = False,
        tir_limit: Optional[float] = float("inf"),
    ) -> None:
        """
        Initializes a photon batch with specified attributes and sets up trackers for simulation.

        The constructor ensures that directional cosines, location coordinates, and weights are appropriately sized
        to match the batch size (`n`). If only one set is provided, it will be repeated to fill the batch.
        If multiple sets are provided, they must match the batch size.

        Parameters:
        :param wavelength: A single wavelength or an iterable of wavelengths for the photon batch.
        :type wavelength: Union[float, Iterable[float]]
        :param n: The number of photons in the batch. Defaults to 0.
        :type n: int, optional
        :param system: The system containing the photon batch. Defaults to None.
        :type system: Optional[System], optional
        :param directional_cosines: The directional cosines of the photon batch. Can be a single set or an iterable
            matching the batch size. Defaults to (0, 0, 1).
        :type directional_cosines: Iterable[float], optional
        :param location_coordinates: The location coordinates of the photon batch. Can be a single set or an iterable
            matching the batch size. Defaults to (0, 0, 0).
        :type location_coordinates: Iterable[float], optional
        :param weights: The weights of the photon batch. Can be a single value or an iterable matching the batch size. Defaults to 1.
        :type weights: Union[float, Iterable[float]], optional
        :param russian_roulette_constant: The constant used for photon survival in the Russian roulette process. Defaults to 20.
        :type russian_roulette_constant: float, optional
        :param recurse: A flag indicating whether recursion is allowed in the simulation. Defaults to True.
        :type recurse: bool, optional
        :param recursion_depth: The current depth of recursion. Defaults to 0.
        :type recursion_depth: int, optional
        :param recursion_limit: The limit for recursion depth. Defaults to 100.
        :type recursion_limit: float, optional
        :param throw_recursion_error: A flag indicating whether to throw an error when the recursion limit is exceeded. Defaults to True.
        :type throw_recursion_error: bool, optional
        :param keep_secondary_photons: A flag indicating whether to retain secondary photons in the simulation. Defaults to False.
        :type keep_secondary_photons: bool, optional
        :param tir_limit: The limit for total internal reflection (TIR). Defaults to infinity.
        :type tir_limit: float, optional

        Initializes the photon state, sets up directional cosines, location coordinates, and weights for the photon batch.
        Sets up trackers for exit direction, exit location, exit weights, and secondary photons. Initializes photon
        tracking history, such as location and weights history, as well as recursion and TIR count trackers.

        """

        # Init photon state
        self.batch_size = batch_size
        self.wavelength = wavelength
        self.system = system
        self.russian_roulette_constant = russian_roulette_constant
        self._medium = None
        self.recurse = recurse
        self.recursion_depth = recursion_depth
        self.recursion_limit = recursion_limit
        self.throw_recursion_error = throw_recursion_error
        self.keep_secondary_photons = keep_secondary_photons
        self.tir_limit = tir_limit

        # Setup batched attributes
        self._directional_cosines = IndexableProperty(
            self._batch_fill(directional_cosines, dtype=np.float64)
        )
        self._location_coordinates = self._batch_fill(
            location_coordinates, dtype=np.float64
        )
        self._weights = self._batch_fill(weights)

        # Init all-false cache register (so medium will be filled in for all)
        self.cache_register = np.repeat(np.False_, self.batch_size, axis=0)

        # Init empty medium cache and at_interface cache
        self._medium = np.empty((self.batch_size,), dtype=object)
        self.at_interface = np.empty((self.batch_size,), dtype=np.bool_)

        # Exit trackers
        self.exit_direction = np.empty_like(self.directional_cosines)
        self.exit_direction[:] = np.nan
        self.exit_location = np.empty_like(self.location_coordinates)
        self.exit_location[:] = np.nan
        self.exit_weights = np.empty(self.batch_size)
        self.exit_weights[:] = np.nan

        self.secondary_photons = []

        # Call setter in case current_photon is DOA
        self.weights = weights

        # Init trackers
        self.location_history = self.location_coordinates[..., np.newaxis].copy()
        self.weights_history = self.weights[..., np.newaxis].copy()
        self.recursed_photons = np.zeros(self.batch_size, dtype=np.uint16)
        self.tir_count = np.zeros(self.batch_size, dtype=np.uint16)
        self.A = 0.0
        self.T = 0.0
        self.R = 0.0

    def _batch_fill(
        self, value: Union[Iterable, NDArray], dtype: Optional[np.dtype] = None
    ) -> NDArray:
        """
        Provides a filled-in array that matches the batch size of the photon.

        This helper method ensures a value is correctly expanded to match the batch size when necessary.

        :param value: The value to assign, either an IndexableProperty or an NDArray. This will be expanded
                      to match the batch size.
        :type value: Union[IndexableProperty, NDArray]
        :return: A numpy array with the correct batch size.
        :rtype: NDArray[np.float64]
        """
        value = np.asarray(value, dtype=dtype)
        # Return already batched inputs
        if value.shape and value.shape[0] == self.batch_size:
            return value

        # Determine singlet size of value (in cases of pre-dimed/unsqueezed arrays, shape[-1] will give singlet shape)
        base_shape = 1 if np.ndim(value) == 0 else np.shape(value)[-1]
        if np.ndim(value) <= 1:
            value = np.repeat(value[np.newaxis, ...], self.batch_size, axis=0)
        elif np.ndim(value) == 2 and np.shape(value)[0] == 1:
            value = np.repeat(value, self.batch_size, axis=0)
        else:
            raise ValueError(
                f"Input is incompatible with batch size. Input must be shape ({base_shape},) or "
                f"({self.batch_size}, {base_shape}) but the input is {np.shape(value)}"
            )
        return value

    @property
    def location_coordinates(self) -> NDArray[float]:
        """
        Retrieves the current value of the photon location coordinates.

        :return: The current value of the location coordinates as an np.ndarray.
        :rtype: NDArray[float]
        """
        return self._location_coordinates

    @location_coordinates.setter
    def location_coordinates(self, location_coordinates: Iterable[float]) -> None:
        """
        Sets the current value of the location coordinates and update the medium cache register for those that have
        change.
        :param location_coordinates:
        :return:
        """

        # Ensure ndarray and fill to batch size
        location_coordinates = self._batch_fill(
            np.asarray(location_coordinates, dtype=np.float64)
        )

        # Determine change status
        unchanged = location_coordinates[:, 2] == self.location_coordinates[:, 2]

        # Set coordinates
        self._location_coordinates = location_coordinates

        # Update cache register
        self.cache_register = np.where(unchanged, self.cache_register, np.False_)

    @property
    def directional_cosines(self) -> IndexableProperty[float]:
        """
        Retrieves the current value of the photon directional cosines.

        :return: The current value of the directional cosines, wrapped in an `IndexableProperty` object.
        :rtype: IndexableProperty[float]
        """
        return self._directional_cosines

    @directional_cosines.setter
    def directional_cosines(self, directional_cosines: Iterable[float]) -> None:
        """
        Setter for the photon directional cosines that ensures the normalization is maintained and updated the cache
        register where the sign of the z-direction changes to trigger medium updates for those photons.

        This setter automatically re-assigns the updated value to an indexable property object,
        which handles normalization both at creation and when setting values using an indexed `__setitem__`.

        :param value: An iterable of float numbers representing the directional cosines of the photon.
        :type value: Iterable[float]
        :return: This method updates the directional cosines and does not return any value.
        :rtype: None
        """
        # Ensure Indexable Property and fill to batch size
        directional_cosines = IndexableProperty(
            self._batch_fill(directional_cosines, dtype=np.float64), normalize=True
        )

        # Determine changed directions
        unchanged = np.sign(directional_cosines[..., 2]) == np.sign(
            self.directional_cosines[:, 2]
        )

        # Set directional cosines
        self._directional_cosines = directional_cosines

        # Update cache register
        self.cache_register = np.where(unchanged, self.cache_register, np.False_)

    def copy(self, **kwargs: Any) -> Photon:
        """
        Creates a deep copy of the Photon object and allows for overwriting specific attributes using **kwargs.

        This method creates a new instance of the Photon object with the same attributes as the original. It can also
        overwrite the values of specified attributes using the keyword arguments passed to it. The tracker attributes
        (`T`, `R`, `A`, `tir_count`, `recursed_photons`) are automatically reset to 0 in all cases.

        :param kwargs: Keyword arguments representing attributes to overwrite in the copied object.
        :type kwargs: dict
        :return: A new Photon object that is a deep copy of the original -- except it points to the same system -- with
            overwritten attributes if provided.
        :rtype: Photon
        """

        new_obj = copy.deepcopy(self)
        new_obj.system = self.system
        # Check for kwarg overwrites
        for key, value in kwargs.items():
            if hasattr(new_obj, key):
                setattr(new_obj, key, value)
            elif hasattr(new_obj, f"_{key}"):
                setattr(new_obj, f"_{key}", value)

        # Reset tracker attributes
        for key in ["T", "R", "A"]:
            setattr(new_obj, key, 0)
        for key in ["tir_count", "recursed_photons"]:
            setattr(new_obj, key, np.zeros(self.batch_size, dtype=np.uint8))
        new_obj.location_history = self.location_coordinates[..., np.newaxis].copy()
        new_obj.weights_history = self.weights[..., np.newaxis].copy()

        return new_obj

    def __repr__(self) -> str:
        """
        Returns a string representation of the Photon object for quick debugging or analysis.

        This method generates a human-readable string that includes the object's attributes and their values,
        providing an easy way to inspect the state of the Photon.

        :return: A string representation of the Photon object.
        :rtype: str
        """
        out = ""
        for key, val in self.__dict__.items():
            out += f"{key.strip('_')}: {val}\n"
        return out

    def simulate(self) -> None:
        """
        Simulates the behavior of a batch of photons until all photons are terminated and updates the object's
        attributes accordingly.

        The simulation involves the following steps:
        - **Absorption**: Photons interact with the medium and may be absorbed. (See `self.absorb`).
        - **Movement**: Photons move within the medium. (See `self.move`).
        - **Scattering**: Photons may scatter as they travel through the medium. (See `self.scatter`).

        The exact behavior of these events depends on the optical properties of the `Medium` object the photon is in,
        which is determined through queries to the `System` that contains the photon, i.e., `self.system`.

        :return: This method runs the simulation and does not return any value.
        :rtype: None
        """

        if not self.system is not None:
            raise RuntimeError(
                "Photon must be in an Optical System object to simulate."
            )
        if self.recursion_depth >= self.recursion_limit:
            raise RecursionError(
                "Maximum photon recursion limit reached. Recursion depth limit can be increased with the "
                "recursion_limit attribute.\n"
                "To switch this error off and throw a warning instead, set throw_recursion_error to FALSE. This "
                "will simulate the photon to the limit without recursion, rather than throwing an error."
            )
        while not self.is_terminated:
            self.absorb()
            self.move()
            self.scatter()

    @property
    def weights(self) -> NDArray[float]:
        """
        Retrieves the current weights of the photons.
        Retrieves the current weights of the photons.

        :return: An array of float values representing the photon weights.
        :rtype: NDArray[float]
        """
        return self._weights

    @weights.setter
    def weights(self, weights: Union[float, Iterable[float]]) -> None:
        """
        Sets the photon weights and applies Russian roulette if the weights falls below the threshold of 0.005.

        If the weights is below 0.005, the `russian_roulette` method is called to determine whether the photon
        survives or is terminated.

        :param weights: The new weights value(s) to set. Can be a single float number or an iterable of float numbers.
        :type weights: Union[float, Iterable[float]]
        :return: This method updates the weights in place and does not return a value.
        :rtype: None
        """
        if isinstance(weights, Real) or np.shape(weights)[0] == 1:
            weights *= np.ones(self.batch_size)
        self._weights = np.where(weights > 0, weights, 0)
        rr_check = (0 < weights) & (weights < 0.005)
        if np.any(rr_check):
            self.russian_roulette(rr_check)

    def russian_roulette(self, mask: NDArray[np.bool_]) -> None:
        """
        Determines photon survival using the Russian roulette technique.

        If a photon survives, its weights is increased by `self.russian_roulette_constant`.
        If it does not survive, its weights is set to `0`, effectively terminating it.

        :param mask: A boolean array where `True` indicates that the corresponding photon is
                     subject to the Russian roulette survival test.
        :type mask: NDArray[np.bool_]
        :return: This method modifies photon weights in place and does not return a value.
        :rtype: None
        """
        survival = np.random.rand(np.count_nonzero(mask)) < (
            1 / self.russian_roulette_constant
        )
        self._weights[mask] = np.where(
            survival, self._weights[mask] * self.russian_roulette_constant, 0
        )

    @property
    def medium(self) -> NDArray[Medium]:
        """
        Determines the current medium for the photons. If the current medium is cached, it is returned from the cache,
        otherwise it is updated by querying the system with the location.

        If the photons are at an interface, the function returns the medium the photon is moving into,
        using `headed_into`. Otherwise, it returns the current medium.

        :return: An array of `Medium` objects representing the current medium for the photons.
        :rtype: NDArray[Medium]
        """
        # Fill in from the cache
        cached = self._medium
        self._medium = np.empty((self.batch_size,), dtype=object)
        self._medium[self.cache_register] = cached[self.cache_register]

        # Update changed photons
        self._medium[~self.cache_register] = self.system.in_medium(
            self.location_coordinates[~self.cache_register]
        )

        # Photons that are at an interface have a tuple returned from system query
        self.at_interface[~self.cache_register] = np.array(
            [isinstance(medium, (tuple, list)) for medium in self._medium]
        )[~self.cache_register]

        # Get the headed_into medium for those photons
        if np.any(self.at_interface):
            self._medium[self.at_interface] = self.headed_into(mediums=self._medium)[
                self.at_interface
            ]

        # Update the cache register to all-true
        self.cache_register[:] = np.True_

        return self._medium

    @property
    def is_terminated(self) -> np.bool_:
        """
        Checks if there are still photons in the batch to simulate.

        This property returns `True` if all photons have been terminated (i.e., they have exited the simulation
        or met termination criteria). Otherwise, it returns `False`.

        :return: A boolean value indicating whether the simulation has completed for all photons.
        :rtype: np.bool_
        """

        self._is_terminated = np.all(
            [
                medium is self.system.surroundings for medium in self.medium
            ]  # Outside the system
            | (self.weights <= 0.0)  # Fully absorbed
            | np.any(
                np.isinf(self.location_coordinates)
                | np.isnan(self.location_coordinates),
                axis=1,
            )  # At infinite
        )
        return self._is_terminated

    def headed_into(
        self, mediums: Optional[NDArray[Union[Medium, Tuple[Medium, Medium]]]] = None
    ) -> NDArray[Medium]:
        """
        Determines which medium a photon is headed into, particularly when at an interface.

        If `mediums` is provided, the function uses it instead of re-querying `self.system.in_mediums` to determine the
        current state. In cases where `mediums` contains tuples `(Medium, Medium)`, the function selects the appropriate
        medium based on direction:
        - The first element (index `0`) represents the negative-direction medium.
        - The second element (index `1`) represents the positive-direction medium.

        If `mediums` is not a tuple, the function simply returns the same medium.

        :param mediums: An optional array of mediums. Each element can be:
            - A `Medium` object, indicating a single medium.
            - A tuple `(Medium, Medium)`, representing a negative and positive medium pair at an interface.
            If not provided, the function queries `self.system.in_mediums` instead.
        :type mediums: NDArray[Union[Medium, Tuple[Medium, Medium]]], optional
        :return: An array of `Medium` objects representing the medium the photon is headed into.
        :rtype: NDArray[Medium]
        """

        mediums = (
            self.system.in_medium(self.location_coordinates)
            if mediums is None
            else mediums
        )
        in_mask = np.array([isinstance(medium, (tuple, list)) for medium in mediums])
        neg_medium = [
            medium[0] if in_ else np.nan for in_, medium in zip(in_mask, mediums)
        ]
        pos_medium = [
            medium[1] if in_ else np.nan for in_, medium in zip(in_mask, mediums)
        ]
        headed_into = np.where(
            ~in_mask,
            mediums,
            np.where(self.directional_cosines[:, 2] < 0, neg_medium, pos_medium),
        )

        return headed_into

    # TODO: Add fluorescence support (dont forget to consider quantum yield < 1)
    def absorb(self) -> None:
        """
        Decrements the weights of the photon batch according to the albedo of the current medium.

        This method simulates the absorption of photons in the current medium. The weights of each photon is
        reduced based on the albedo, which represents the fraction of light that is reflected. The remaining
        weights is used to continue the photon simulation.

        The specific absorption process depends on the optical properties of the medium the photon is currently in.

        :return: None. The photon batch's weights are updated in place.
        :rtype: None
        """

        absorbed_weights = self.weights * np.array(
            [medium.albedo_at(self.wavelength) for medium in self.medium]
        )
        self.A += np.sum(absorbed_weights)
        self.weights = self.weights - absorbed_weights

    def move(self, step: Union[float, Iterable[float]] = None) -> None:
        """
        Moves the photon one step in its current direction.

        The size of the step can be either provided directly, or it can be determined by the mean free path of the
        current medium, which is calculated from the transport coefficient (mu_t). If the photon is in a medium with
        mu_t = 0 (i.e., no scattering or absorption), the photon will automatically advance to the next interface along
        its direction.

        If a step size is provided, the photon moves in the direction specified by its directional cosines until it
        either completes the step or hits an interface. If the full step would cross an interface, only the portion of
        the step before the interface is executed. At the interface, the photon is refracted and reflected according to
        the media properties, and its weights is updated (decremented) based on the interaction.

        :param step: The distance the photon should move. If not provided, the step size is sampled based on the mean
        free path.
        :type step: Union[float, Iterable[float]], optional
        :return: None. The photon’s state (location, weights) is updated in place.
        :rtype: None
        """

        # Get current state
        mu_t = np.array([medium.mu_t_at(self.wavelength) for medium in self.medium])
        dir_cos = self.directional_cosines
        loc = self.location_coordinates

        # If scattering occurs, step is sampled from the distribution
        if step is None:
            step = np.where(
                mu_t > 0, -np.log(np.random.rand(self.batch_size)) / mu_t, float("inf")
            )
            step = np.where(self.weights > 0, step, 0)
        step = self._batch_fill(step)
        new_loc = loc + step[:, np.newaxis] * dir_cos

        # Determine which photons cross an interfaces
        interface, plane = self.system.interface_crossed(loc, new_loc)
        crossed = (~np.isnan(plane)) if plane is not None else False
        move_to_interface = False
        if np.any(crossed):
            interface_steps = np.where(
                crossed, (plane - loc[:, 2]) / dir_cos[:, 2], float("inf")
            )

            # Find photons that should move to the interfaces instead
            move_to_interface = (interface_steps < step) & crossed
            new_loc[move_to_interface] = (
                loc[move_to_interface]
                + interface_steps[move_to_interface, np.newaxis]
                * dir_cos[move_to_interface]
            )

        # Update location
        self.location_coordinates = new_loc

        # Reflect/refract photons at interfaces
        if np.any(move_to_interface):
            self.reflect_refract(interface, move_to_interface)

        # Update history for all photons
        self.location_history = np.append(
            self.location_history, new_loc[..., np.newaxis], axis=2
        )
        self.weights_history = np.append(
            self.weights_history, self.weights[..., np.newaxis], axis=1
        )

        # Check if any new photons exited
        exit_mask = np.array(
            [medium is self.system.surroundings for medium in self.headed_into()]
        ) & (self.weights > 0)

        if np.any(exit_mask):
            self.exit_location[exit_mask] = self.location_history[exit_mask, ..., -1]
            self.exit_weights[exit_mask] = self.weights_history[exit_mask, -1]

            # Check if any exited photons hit a detector
            detector_mask = exit_mask & (
                self.exit_location[:, 2] == self.system.detector_location
            )
            if np.any(detector_mask):
                self.system.detector(self, exit_mask)

            # Handle reflection or transmission
            self.R += np.sum(
                np.where(
                    exit_mask & (self.directional_cosines[:, 2] < 0), self.weights, 0
                )
            )
            self.T += np.sum(
                np.where(
                    exit_mask & (self.directional_cosines[:, 2] > 0), self.weights, 0
                )
            )

            # Terminate exited photons
            self.weights[exit_mask] = 0

    def reflect_refract(
        self,
        interfaces: NDArray[object[Tuple[Medium, Medium], Tuple[None]]],
        mask: NDArray[np.bool_],
    ) -> None:
        """
        Computes photon reflection and refraction at interfaces.

        This method determines the weight of photon reflection and updates the tracker's direction accordingly.
        If reflection occurs, it either updates the photon direction or spawns secondary photons with adjusted weights
        to continue the recursive simulation. The photon's direction is updated based on the interface's refractive
        values. This operation is performed only on photons specified by the input mask.

        :param interfaces: NumPy array containing interface definitions.
                           Each element is a tuple of (Medium, Medium) for a valid interface or (None,) if no interface
                           is present.
        :type interfaces: np.ndarray[object]
        :param mask: Boolean mask indicating which photons to process.
        :type mask: NDArray[np.bool_]
        :return: None. The photon's direction and weight are updated along with relevant trackers,
                 and secondary photons are spawned if necessary.
        """

        # Get incidence state
        mu_x, mu_y, mu_z_i = self.directional_cosines[mask].T
        mu_z_t = np.zeros_like(mu_z_i)
        n1 = np.array(
            [
                (
                    interface[0].n
                    if iterable(interface) and len(interface) == 2
                    else np.nan
                )
                for interface in interfaces
            ]
        )[mask]
        n2 = np.array(
            [
                (
                    interface[1].n
                    if iterable(interface) and len(interface) == 2
                    else np.nan
                )
                for interface in interfaces
            ]
        )[mask]

        # Calculate refraction
        sin_theta_t = n1 / n2 * np.sqrt(1 - (mu_z_i**2))

        # TIR
        tir_mask = sin_theta_t > 1
        mu_z_t[tir_mask] = -mu_z_i[tir_mask]
        self.tir_count[mask] = np.where(
            tir_mask, self.tir_count[mask] + 1, self.tir_count[mask]
        )
        stop_tir = self.tir_count[mask] > self.tir_limit
        self.A += np.sum(np.where(stop_tir, self.weights[mask], 0))
        self.weights[mask] = np.where(stop_tir, 0, self.weights[mask])

        # Snell's + Fresnel's Law
        refract_mask = ~tir_mask
        mu_z_t_masked = np.sqrt(1 - (sin_theta_t[refract_mask] ** 2))

        # Extract only the masked refractive values for masked values
        n1_masked = n1[refract_mask]
        n2_masked = n2[refract_mask]
        abs_mu_z_i = np.abs(mu_z_i[refract_mask])

        rs = (
            np.abs(
                ((n1_masked * abs_mu_z_i) - (n2_masked * mu_z_t_masked))
                / ((n1_masked * abs_mu_z_i) + (n2_masked * mu_z_t_masked))
            )
            ** 2
        )
        rp = (
            np.abs(
                ((n2_masked * mu_z_t_masked) - (n1_masked * abs_mu_z_i))
                / ((n1_masked * abs_mu_z_i) + (n2_masked * mu_z_t_masked))
            )
            ** 2
        )
        specular_reflection = 0.5 * (rs + rp) * self.weights[mask][refract_mask]

        mu_z_t[refract_mask] = mu_z_t_masked * np.sign(
            mu_z_i[refract_mask]
        )  # Ensure correct sign

        # Updated for transmitted portion
        mu_x[refract_mask] *= n1_masked / n2_masked
        mu_y[refract_mask] *= n1_masked / n2_masked

        if self.recurse:
            # Setup new photon attributes
            batch_size = np.sum(refract_mask).item()
            dir_cos = (
                np.array([[1, 1, -1]]) * self.directional_cosines[mask][refract_mask]
            )  # Flipped for reflection
            weights = specular_reflection
            loc_cor = self.location_coordinates[mask][refract_mask]
            rec_dep = self.recursion_depth + 1
            secondary_photons = Photon(
                self.wavelength,
                system=self.system,
                batch_size=batch_size,
                location_coordinates=loc_cor,
                directional_cosines=dir_cos,
                weights=weights,
                recursion_depth=rec_dep,
            )
            secondary_photons._medium = self._medium[mask][refract_mask]
            secondary_photons.cache_register = self.cache_register[mask][refract_mask]
            secondary_photons.at_interface = self.at_interface[mask][refract_mask]
            try:
                secondary_photons.simulate()
            except RecursionError as e:
                if self.throw_recursion_error:
                    raise e
                else:
                    warnings.warn(str(e), RuntimeWarning)
            finally:
                self.R += secondary_photons.R
                self.T += secondary_photons.T
                self.A += secondary_photons.A
                if self.keep_secondary_photons:
                    self.secondary_photons.append(secondary_photons)
                    self.secondary_photons += secondary_photons.secondary_photons

        else:
            # If the reflected fraction will be reflected out, add it to reflected count, Else add it to transmitted
            reflected_out = mu_z_i[refract_mask] > 0
            transmitted_out = mu_z_i[refract_mask] < 0
            self.R += np.sum(specular_reflection * reflected_out)
            self.T += np.sum(specular_reflection * transmitted_out)

        w_temp = self.weights[mask].copy()
        w_temp[refract_mask] -= specular_reflection
        self.weights[mask] = w_temp

        # Send to setter for normalization
        self.directional_cosines[mask] = np.column_stack((mu_x, mu_y, mu_z_t))

    def scatter(
        self,
        theta_phi: Optional[
            Union[Iterable[float, float], Iterable[Iterable[float, float]]]
        ] = None,
    ):
        """
        This method updates the direction of the photon members where they are in scattering media, but not at an
        interface.

        :param theta_phi: Angles to update direction with. Optional.
        :type theta_phi: Union[Iterable[float, float], Iterable[Iterable[float, float]]]
        :return: None. Updates photon directions where the photon is in scattering media (but not at an interface)
        """
        # Early break if all are at an interface or in non-scattering medium
        g = np.array(
            [medium.g for medium in self.medium]
        )  # (also forces reset of interface cache where necessary)
        if np.all(self.at_interface) or np.all(g == 1):
            return

        # Placeholders for angle samples
        theta = np.zeros(self.batch_size, dtype=np.float64)
        cosine_theta = np.zeros_like(theta, dtype=np.float64)
        if theta_phi is None:
            # Sample random scattering angles from distribution
            [xi, zeta] = np.random.rand(self.batch_size, 2).T

            # For non-zero g
            non_zero_g_mask = g != 0
            lead_coeff = 1 / (2 * g[non_zero_g_mask])
            term_2 = g[non_zero_g_mask] ** 2
            term_3 = (1 - term_2) / (
                1 - g[non_zero_g_mask] + (2 * g[non_zero_g_mask] * xi[non_zero_g_mask])
            )
            cosine_theta[non_zero_g_mask] = lead_coeff * (1 + term_2 - term_3)

            # For g=0
            cosine_theta[~non_zero_g_mask] = 1 - (2 * xi[~non_zero_g_mask])

            theta = np.arccos(cosine_theta)
            phi = 2 * np.pi * zeta
        else:
            theta, phi = (
                theta_phi if isinstance(theta_phi, (tuple, list)) else zip(theta_phi)
            )
        # Update direction cosines
        mu_x, mu_y, mu_z = self.directional_cosines.T
        new_directional_cosines = np.zeros((self.batch_size, 3), dtype=np.float64)

        # For near-vertical photons (simplify for stability)
        vertical = np.abs(mu_z) > 0.999
        new_directional_cosines[vertical, 0] = np.sin(theta[vertical]) * np.cos(
            phi[vertical]
        )
        new_directional_cosines[vertical, 1] = (
            np.sign(mu_z[vertical]) * np.sin(theta[vertical]) * np.sin(phi[vertical])
        )
        new_directional_cosines[vertical, 2] = np.sign(mu_z[vertical]) * np.cos(
            theta[vertical]
        )

        # For all others
        nonvertical = ~vertical
        deno = np.sqrt(1 - (mu_z[nonvertical] ** 2))

        # mu_x updated
        numr = (mu_x[nonvertical] * mu_z[nonvertical] * np.cos(phi[nonvertical])) - (
            mu_y[nonvertical] * np.sin(phi[nonvertical])
        )

        new_directional_cosines[nonvertical, 0] = (
            np.sin(theta[nonvertical]) * (numr / deno)
        ) + (mu_x[nonvertical] * np.cos(theta[nonvertical]))

        # mu_y update
        numr = (mu_y[nonvertical] * mu_z[nonvertical] * np.cos(phi[nonvertical])) + (
            mu_x[nonvertical] * np.sin(phi[nonvertical])
        )

        new_directional_cosines[nonvertical, 1] = (
            np.sin(theta[nonvertical]) * (numr / deno)
        ) + (mu_y[nonvertical] * np.cos(theta[nonvertical]))

        # mu_z update
        new_directional_cosines[nonvertical, 2] = -(
            np.sin(theta[nonvertical]) * np.cos(phi[nonvertical]) * deno
        ) + (mu_z[nonvertical] * np.cos(theta[nonvertical]))

        # Update directional cosines with new direciton (done at once for normalization consistency)
        self.directional_cosines[~self.at_interface] = new_directional_cosines[
            ~self.at_interface
        ]

    def plot_path(
        self,
        project_onto: Optional[str] = None,
        axes: Optional[plt.Axes] = None,
        ignore_outside: bool = True,
    ) -> None:
        """
        Visualizes the photon location histories as a 2D or 3D plot.

        - If `project_onto` is specified, the path is projected onto a 2D plane.
          Accepted values: 'xy', 'xz', 'yz'.
        - If `project_onto` is None, a 3D plot is generated.
        - If `axes` is provided, the plot is drawn on the given `matplotlib.Axes` object.
          Otherwise, a new figure and axes are created.
        - If `ignore_outside` is True (default), only locations within the system are plotted,
          and steps after exit are truncated.

        :param project_onto: Plane onto which the path is projected ('xy', 'xz', 'yz'), or None for 3D.
        :type project_onto: Optional[str]
        :param axes: Matplotlib Axes object for plotting, or None to create new axes.
        :type axes: Optional[plt.Axes]
        :param ignore_outside: Whether to ignore steps after exiting the system (default: True).
        :type ignore_outside: bool
        :return: None
        """

        project_onto = ["xz", "yz", "xy"] if project_onto == "all" else project_onto
        project_onto = (
            [project_onto]
            if isinstance(project_onto, str) or project_onto is None
            else project_onto
        )
        batch_size, _, steps = (
            self.location_history.shape
        )  # Expect shape (batch, 3, steps)

        # Boundaries for filtering
        z_min, z_max = self.system.boundaries[0], self.system.boundaries[-1]
        inside = (
            (
                (self.location_history[:, 2] >= z_min)
                & (self.location_history[:, 2] <= z_max)
            )
            if ignore_outside
            else True
        )

        fig = (
            plt.figure(figsize=(8 * len(project_onto), 8))
            if not plt.get_fignums()
            else plt.gcf()
        )
        if project_onto[0]:
            axes = (
                [
                    fig.add_subplot(1, len(project_onto), i + 1)
                    for i in range(len(project_onto))
                ]
                if axes is None
                else axes
            )
            for ax, projection in zip(axes, project_onto):
                for i in range(batch_size):
                    x, y = (
                        self.location_history[i, "xyz".index(projection[0])],
                        self.location_history[i, "xyz".index(projection[1])],
                    )
                    ax.plot(x[inside[i]], y[inside[i]], label=f"Photon {i + 1}")
                ax.set_title(f"Projected onto {projection}-plane")
                ax.set_xlabel(f"Photon Displacement in {projection[0]}-direction (cm)")
                ax.set_ylabel(f"Photon Displacement in {projection[1]}-direction (cm)")
                if projection[1] == "z" and not ax.yaxis_inverted():
                    ax.invert_yaxis()
                if projection[0] == "z" and not ax.xaxis_inverted():
                    ax.invert_xaxis()
        else:
            axes = fig.add_subplot(projection="3d") if axes is None else axes
            for i in range(batch_size):
                x, y, z = (
                    self.location_history[i, 0],
                    self.location_history[i, 1],
                    self.location_history[i, 2],
                )
                axes.plot(
                    x[inside[i]], y[inside[i]], z[inside[i]], label=f"Photon {i + 1}"
                )
            axes.set_title("Photon Paths")
            axes.set_xlabel("Photon Displacement in x-direction (cm)")
            axes.set_ylabel("Photon Displacement in y-direction (cm)")
            axes.set_zlabel("Photon Displacement in z-direction (cm)")
            if not axes.zaxis_inverted():
                axes.invert_zaxis()

        return fig, axes



import functools
from numbers import Real
from typing import Tuple, Union, Callable, Optional, Iterable
from .import_utils import np, iterable, NDArray

# Importable and default hardware specs
# ID: Inner diameter in cm
ID = 0.16443276801785274

# OD: Outer diameter in cm
OD = 0.3205672319821467

# WD: Working distance in cm
WD = 0.2

# THETA: Angle of cone based on geometry
THETA = np.arctan(-OD / WD)  # rad

# NA: Numerical aperture of the optical system
NA = 1.0

# Create random number generator
rng = np.random.default_rng()


# Default "sampler" and "detector" to start photons straight down at origin and detect all reflected
def pencil_beams(n: int) -> Tuple[NDArray[float], NDArray[float]]:
    """
    Generates a pencil beam of `n` photons originating from the origin and directed straight downward.

    :param n: Number of photons to simulate.
    :type n: int
    :return: Tuple of (locations, directional_cosines) arrays.
    :rtype: Tuple[NDArray[float], NDArray[float]]
    """
    return (
        np.repeat(np.array((0, 0, 0))[np.newaxis, ...], n, axis=0),
        np.repeat(np.array((0, 0, 1))[np.newaxis, ...], n, axis=0),
    )


def hollow_cone_beam(
    n: int = 50000,
    r_min: float = ID,
    r_max: float = OD,
    angle_min: float = THETA,
    angle_max: float = THETA,
) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
    """
    Samples photons from a hollow conical beam with specified radial and angular bounds.

    :param n: Number of photons to generate.
    :type n: int
    :param r_min: Minimum radius from beam center (cm).
    :type r_min: float
    :param r_max: Maximum radius from beam center (cm).
    :type r_max: float
    :param angle_min: Minimum injection angle in radians.
    :type angle_min: float
    :param angle_max: Maximum injection angle in radians.
    :type angle_max: float
    :return: Tuple of photon launch locations and directional cosines.
    :rtype: Tuple[NDArray[np.float64], NDArray[np.float64]]
    """
    # Sample angle and radius for starting location
    phi = np.random.uniform(0, 2 * np.pi, n)
    r = np.sqrt(np.random.uniform(r_min**2, r_max**2, n))

    # Create ring (2D coordinates)
    x = r * np.cos(phi)
    y = r * np.sin(phi)
    location = np.vstack((x, y, np.zeros(n))).T  # Stack x, y, and z=0 into shape (n, 3)

    # Sample injection angles directional cosines
    theta = np.random.uniform(angle_min, angle_max, n)

    # Compute directional cosines
    mu_x = np.sin(theta) * np.cos(phi)
    mu_y = np.sin(theta) * np.sin(phi)
    mu_z = np.cos(theta)
    directional_cosines = np.vstack((mu_x, mu_y, mu_z)).T  # Stack into shape (n, 3)

    return location, directional_cosines


# NOTE: This samples the ring _at_ where ID/OD are measured from, presumably a distance WD above the medium of measure.
def create_hollow_cone_beam(
    r_bounds: Union[Real, Tuple[Real, Real]],
    angle_bounds: Union[Real, Tuple[Real, Real]],
) -> Callable:
    """
    Factory function to create a parameterized hollow cone beam sampler.

    :param r_bounds: Either a single max radius or a (min, max) tuple.
    :type r_bounds: Union[Real, Tuple[Real, Real]]
    :param angle_bounds: Either a single angle or a (min, max) tuple in radians.
    :type angle_bounds: Union[Real, Tuple[Real, Real]]
    :return: A callable beam generator with fixed bounds.
    :rtype: Callable
    """
    if not iterable(r_bounds):
        r_max = r_bounds
        r_min = 0
    elif len(r_bounds) == 1:
        r_max = r_bounds[0]
        r_min = 0
    else:
        r_min, r_max = r_bounds

    if not iterable(angle_bounds):
        angle_min = angle_bounds
        angle_max = angle_bounds
    elif len(angle_bounds) == 1:
        angle_min = angle_bounds[0]
        angle_max = angle_bounds[0]
    else:
        angle_min, angle_max = angle_bounds

    return functools.partial(
        hollow_cone_beam,
        r_min=r_min,
        r_max=r_max,
        angle_min=angle_min,
        angle_max=angle_max,
    )


def oblique_beams(
    n: int, beams: Tuple[int, ...] = (0, 1), gamma: float = 75, w: float = 1
):
    """
    Generates obliquely incident beams from a tilted source with circular beam profiles.

    :param n: Number of photons to simulate.
    :type n: int
    :param beams: Tuple indicating axes used for oblique beam injection (0 for x, 1 for y).
    :type beams: Tuple[int, ...]
    :param gamma: Angle in degrees from vertical (z-axis).
    :type gamma: float
    :param w: Beam width in cm.
    :type w: float
    :return: Tuple of photon launch locations and directional cosines.
    :rtype: Tuple[NDArray[float], NDArray[float]]
    """
    c = np.radians(gamma)

    # Sample radial positions within a circular beam profile
    r = w * np.sqrt(np.random.uniform(0, 1, n))  # Uniform distribution in area
    phi = np.random.uniform(0, 2 * np.pi, n)  # Uniform angular distribution

    # Calculate the offsets based on the radial distance and angle
    in_plane_offsets = r * np.cos(phi)
    out_of_plane_offsets = r * np.sin(phi)

    # Calculate incidence point on horizontal plane
    in_plane_incidence = in_plane_offsets + (r * np.sin(c)) - (w * np.sin(c) / 2)
    in_plane_incidence /= np.cos(c)
    out_of_plane_incidence = out_of_plane_offsets  # Out of plane location is unchanged

    # Select beam and input final outputs accordingly (beam0: xz plane; beam1: yz plane)
    beam = np.random.choice(beams, n) if len(beams) > 1 else beams[0]

    location_coordinates = np.zeros((n, 3))
    directional_cosines = np.cos(c) * np.ones((n, 3))
    for i in beams:
        location_coordinates[:, i] = np.where(
            beam == i, in_plane_incidence, out_of_plane_incidence
        )
        directional_cosines[:, i] = np.where(beam == i, np.sin(c), 0)

    return location_coordinates, directional_cosines


def create_oblique_beams(
    beams: Tuple[int, ...] = (0, 1), gamma: float = 75, w: float = 1
):
    """
    Factory function to create an oblique beam sampler.

    This wraps the `oblique_beams` function with preset parameters for use in multiprocessing contexts.

    :param beams: Tuple indicating beam orientation axes.
    :type beams: Tuple[int, ...]
    :param gamma: Angle in degrees from the z-axis.
    :type gamma: float
    :param w: Width of the beam (cm).
    :type w: float
    :return: Callable beam generator with frozen parameters.
    :rtype: Callable
    """
    return functools.partial(oblique_beams, beams=beams, gamma=gamma, w=w)


# Stand-in detector for all queried
def total_acceptor(
    x: float, y: float, mu_z: Optional[float] = None
) -> NDArray[np.bool_]:
    """
    Detector function that accepts all reflected photons unconditionally.

    :param x: x-position(s) of the photon.
    :type x: float
    :param y: y-position(s) of the photon.
    :type y: float
    :param mu_z: Optional z-direction cosine(s); ignored.
    :type mu_z: Optional[float]
    :return: Boolean mask indicating all photons are accepted.
    :rtype: NDArray[np.bool_]
    """
    return np.full_like(x, fill_value=True, dtype=np.bool_)


def cone_of_acceptance(
    x: Union[Real, Iterable[Real]],
    y: Union[Real, Iterable[Real]],
    mu_z: Union[Real, Iterable[Real]] = None,
    na: float = NA,
    n: float = 1.33,
    r: float = ID,
) -> NDArray[np.bool_]:
    """
    Detector function that accepts photons based on angle and radial distance.

    Accepts photons within a specified numerical aperture and radial extent.

    :param x: x-coordinate(s) of the photon.
    :type x: Union[Real, Iterable[Real]]
    :param y: y-coordinate(s) of the photon.
    :type y: Union[Real, Iterable[Real]]
    :param mu_z: z-direction cosine(s) for angular filtering.
    :type mu_z: Union[Real, Iterable[Real]]
    :param na: Numerical aperture.
    :type na: float
    :param n: Refractive index of the medium.
    :type n: float
    :param r: Maximum radial acceptance.
    :type r: float
    :return: Boolean mask indicating which photons are accepted.
    :rtype: NDArray[np.bool_]
    """
    x = np.array(x)
    y = np.array(y)
    if mu_z is not None:
        theta_max = np.arcsin(na / n)
        mu_z_max = np.cos(theta_max)
        too_steep = np.array(mu_z) > mu_z_max
    else:
        too_steep = False
    r_test = np.sqrt(x**2 + y**2)
    outside = r_test > r
    return ~too_steep & ~outside


def create_cone_of_acceptance(
        r: Real,
        na: Real = NA,
        n: Real = 1.33
) -> Callable[[Union[Real, Iterable[Real]], Union[Real, Iterable[Real]], ...], NDArray[np.bool_]]:
    """
    Factory function to create a cone-of-acceptance detector function.

    :param r: Radial acceptance limit.
    :type r: Real
    :param na: Numerical aperture.
    :type na: Real
    :param n: Refractive index.
    :type n: Real
    :return: Callable detector function.
    :rtype: Callable
    """
    return functools.partial(cone_of_acceptance, r=r, na=na, n=n)



import importlib.resources
import pandas as pd
import numpy as np

with importlib.resources.open_text(
    "photon_canon.data", "sample_food_coloring_absorption.csv"
) as f:
    df = pd.read_csv(f)
raw_wl, a_red, a_yellow, a_green, a_blue = (
    df["Wavelength"],
    df["red_45ul"],
    df["yellow_45ul"],
    df["green_45ul"],
    df["blue_45ul"],
)

# Original dilution was 45 uL of stock (5uL dye in 3mL water) into 3.4 mL of water
s = (
    45 / 1000 * 5 / 3 / 3.4  # uL -> mL  # stock concentration  # cuvette volume
)  # concentration measured in uL / mL
s = 1 / s  # invert to scale to 1uL / mL

# Normalize units to 1 uL / mL
a_red *= s
a_yellow *= s
a_green *= s
a_blue *= s

# Load pre-mixed validation stocks.
# Stock A: 0.125 mL / mL red, 0.175 mL / mL green in water
# Stock B: 0.600 mL / mL yellow, 0.050 mL / mL blue in water
# Stocks were diluted down for absorbance measurements to 10 uL/mL stock
with importlib.resources.open_text(
    "photon_canon.data", "validation_stock_absorbance.csv"
) as f:
    df = pd.read_csv(f)
premix_wl, a_stock_a, a_stock_b = df["Wavelength"], df["stock_a"], df["stock_b"]

# Normalize to 1 uL / mL
a_stock_a /= 10
a_stock_b /= 10


def make_mix(
    wavelengths: np.ndarray = None,
    *,
    red: float = 0,
    yellow: float = 0,
    green: float = 0,
    blue: float = 0,
    stock_a: float = 0,
    stock_b: float = 0
) -> np.ndarray[float]:

    # Raw colors
    if wavelengths is None:
        wl = raw_wl
    else:
        wl = wavelengths

    mask = raw_wl.isin(wl)
    a_r = a_red[mask].values
    a_y = a_yellow[mask].values
    a_g = a_green[mask].values
    a_b = a_blue[mask].values

    # Premixed
    if wavelengths is None:
        wl = premix_wl

    mask = premix_wl.isin(wl)
    a_sa = a_stock_a[mask].values
    a_sb = a_stock_b[mask].values

    return np.array(
        red * a_r
        + yellow * a_y
        + green * a_g
        + blue * a_b
        + stock_a * a_sa
        + stock_b * a_sb
    )



from photon_canon import Medium, System, Detector, Illumination
from photon_canon.hardware import create_oblique_beams
from photon_canon.import_utils import np
from photon_canon.lut import generate_lut


def main():
    # Init parameter sets
    mu_s_array = np.arange(0, 101, 1)
    mu_a_array = np.arange(1, 102, 1)
    g_array = [0.9]
    d = 0.2
    n = 100000
    tissue_n = 1.33
    surroundings_n = 1
    recurse = False
    wl0 = 650

    # Make water medium
    di_water = Medium(n=1.33, mu_s=0, mu_a=0, g=0, desc="di water", ref_wavelength=wl0)
    glass = Medium(n=1.523, mu_s=0, mu_a=0, g=0, desc="glass", ref_wavelength=wl0)

    # Create an illuminator
    lamp = Illumination(create_oblique_beams((0, 1), 60, 1.5))

    # Start the system
    system = System(
        di_water,
        0.2,  # 1mm
        glass,
        0.017,  # 0.17mm
        surrounding_n=surroundings_n,
        illuminator=lamp,
    )
    tissue = Medium(
        n=tissue_n, mu_s=0, mu_a=0, g=1, desc="tissue"
    )  # Placeholder to update at iteration
    system.add(tissue, d)

    # Generate a photon object (either directly or through the system illumination)
    photon = system.beam(batch_size=n, recurse=recurse)
    generate_lut(
        system,
        tissue,
        {"mu_s": mu_s_array, "mu_a": mu_a_array, "g": g_array},
        photon,
        pbar=True,
        num_workers=1,
    )


if __name__ == "__main__":
    main()



try:
    import cupy as np
    from cupy import iterable
    from cupy.typing import NDArray
    from cupyx.scipy.interpolate import RegularGridInterpolator, interp1d

    if not np.is_available():
        raise RuntimeError("CUDA not available; reverting to NumPy.")
except (ImportError, RuntimeError):
    import numpy as np
    from numpy import iterable
    from numpy.typing import NDArray
    from scipy.interpolate import RegularGridInterpolator, interp1d

    np.is_available = lambda: False  # Mock the `is_available` method for consistenc



import sqlite3
from enum import Enum
from numbers import Real
from pathlib import Path
from typing import Union, Iterable

from ..import_utils import np

from .. import System

# Set up simulation database
db_dir = Path.home() / ".photon_canon"
db_dir.mkdir(parents=True, exist_ok=True)
db_path = db_dir / "lut.db"
con = sqlite3.connect(db_path)
c = con.cursor()


class Dimensions(str, Enum):
    MU_S = "mu_s"
    MU_A = "mu_a"
    G = "g"


class ListPortion(Enum):
    HEAD = slice(None, 5)
    TAIL = slice(-5, None)
    ALL = slice(None, None)


class classOrInstanceMethod(object):
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, owner):

        def func(*args, **kwargs):
            return self.func(instance, owner, *args, **kwargs)

        return func


def add_metadata(n=None, recursive=False, detector=None) -> int:
    # Parse to get metadata
    c.execute(
        """
    CREATE TABLE IF NOT EXISTS mclut_simulations (
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        photon_count INTEGER NOT NULL,
        recursive BOOLEAN DEFAULT FALSE,
        detector BOOLEAN DEFAULT FALSE,
        detector_description TEXT DEFAULT ''
        )
        """
    )

    c.execute(
        f"""
    INSERT INTO mclut_simulations (
    photon_count, recursive, detector, detector_description
    ) VALUES (?, ?, ?, ?)""",
        (
            n,
            recursive,
            detector is not None,
            detector.desc if detector is not None else "",
        ),
    )
    con.commit()

    return c.lastrowid


def add_system_data(simulation_id: int, system: System) -> None:
    c.execute(
        """
    CREATE TABLE IF NOT EXISTS fixed_layers (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    stack_order INTEGER NOT NULL,
    layer TEXT NOT NULL,
    mu_s REAL NOT NULL,
    mu_a REAL NOT NULL,
    g REAL NOT NULL,
    thickness REAL NOT NULL,
    ref_wavelength REAL NOT NULL,
    simulation_id INTEGER NOT NULL,
    FOREIGN KEY (simulation_id) REFERENCES mclut_simulations(id)
    )
    """
    )

    # Generate fixed layer details for table
    fixed_layers = []
    for i, (bound, layer) in enumerate(system.stack.items()):
        fixed_layers.append(
            (
                int(i),
                layer.desc,
                float(layer.mu_s_at(layer.ref_wavelength)),
                float(layer.mu_a_at(layer.ref_wavelength)),
                float(layer.g),
                float(bound[1] - bound[0]),
                float(layer.ref_wavelength),
                int(simulation_id),
            )
        )

    # Add details to table
    c.executemany(
        f"""
    INSERT INTO fixed_layers (
        stack_order, layer, mu_s, mu_a, g, thickness, ref_wavelength, simulation_id
        ) VALUES (?, ?, ?, ?, ?, ?, ?, ?)""",
        fixed_layers,
    )
    con.commit()


def add_simulation_result(
    simulation_id: int,
    mu_s: Union[Real, Iterable[Real]],
    mu_a: Union[Real, Iterable[Real]],
    g: Union[Real, Iterable[Real]],
    depth: Union[Real, Iterable[Real]],
    output: Union[Real, Iterable[Real]],
) -> None:
    # Ensure all are iterable arays
    arrays = [np.array(val) for val in [mu_s, mu_a, g, depth, output, simulation_id]]
    iters = [arr.ndim == 1 for arr in arrays]
    shapes = [arrays[idx].shape for idx, itr in enumerate(iters) if itr]

    # Check that sizes are compatible
    if shapes and not np.all(shapes == shapes[0]):
        raise ValueError("Iterables must have the same shape")
    elif not shapes:
        shapes = [[1]]

    # Expand non-iterables to match and make all floats
    for i, itr in enumerate(iters):
        if not itr:
            arrays[i] = np.repeat([float(arrays[i])], shapes[0][0])
        else:
            arrays[i] = np.array([float(val) for val in arrays[i]])

    # Put into tuples for executemany
    data_tuples = tuple(zip(*arrays))

    # Add table (if not exists)
    c.execute(
        """
     CREATE TABLE IF NOT EXISTS mclut (
         id INTEGER PRIMARY KEY AUTOINCREMENT,
         mu_s REAL NOT NULL,
         mu_a REAL NOT NULL,
         g REAL NOT NULL,
         depth REAL NOT NULL,
         output REAL NOT NULL,
         simulation_id INTEGER NOT NULL,
         FOREIGN KEY (simulation_id) REFERENCES mclut_simulations(id)
         )
         """
    )

    # Add results to db
    c.executemany(
        f"""
                INSERT INTO mclut (
                mu_s, mu_a, g, depth, output, simulation_id
                ) VALUES (?, ?, ?, ?, ?, ?)""",
        data_tuples,
    )

    con.commit()



import itertools
import logging
import multiprocessing as mp
from numbers import Real
from typing import Optional, List, Union, Tuple

from pydantic import PrivateAttr, BaseModel, ConfigDict
from tqdm import tqdm

from ..import_utils import np, NDArray, RegularGridInterpolator
import pandas as pd

from ..optics import System, Medium
from ..optics import Photon

from ..lut.utils import (
    add_metadata,
    add_system_data,
    add_simulation_result,
    Dimensions,
    ListPortion,
    classOrInstanceMethod,
)
from ..utils import latest_simulation_id, CON

c = CON.cursor()


def single_sim(
    args: Tuple[
        System, Medium, Photon, list[str, ...], list[float, ...], Optional[str]
    ],
):
    """Worker function to simulate a single case"""
    system, variable, photon, keys, values, output = args
    keys_values = dict(zip(keys, values))
    variable.set(**keys_values)

    # Reset and simulate
    local_photon = photon.copy()
    if system.detector is not None:
        system.detector.reset()
    local_photon.simulate()

    # Get the output
    if system.detector is not None:
        target = system.detector.n_detected
    elif output is not None:
        target = getattr(photon, output, None)
    else:
        target = getattr(photon, "R", None)

    # Update LUT
    for bound, layer in system.stack.items():
        if layer == variable:
            depth = bound[1] - bound[0]
            break

    return variable.mu_s, variable.mu_a, variable.g, depth, target


def generate_lut(
    system: System,
    variable: Medium,
    arrays: dict[str, NDArray],
    photon: Photon,
    pbar: bool = False,
    output: Optional[str] = None,
    num_workers: int = None,
) -> int:
    """
    Simulates photon transport for different optical properties to generate a lookup table (LUT).

    This function iterates over the input arrays, modifying the optical properties of `variable`
    (a `Medium` object) accordingly. A set of `Photon` objects is then simulated within the `system`.

    :param output: Determines what output measure to add to the LUT. Default total reflectance.
    :param system: The optical system in which the photons are simulated.
    :type system: System
    :param variable: The medium whose properties are varied in the LUT generation.
    :type variable: Medium
    :param arrays: Dictionary mapping property names to NumPy arrays containing values to iterate over.
    :type arrays: dict[str, np.ndarray]
    :param photon: A reference `Photon` object used to initialize the simulated photons.
    :type photon: Photon
    :param pbar: Whether to show a progress bar.
    :type pbar: bool
    :param num_workers: Number of parallel processes to simulate photons.
    :return: The ID of the generated LUT of simulation.
    :rtype: int
    """

    # Add LUT metadata to db
    simulation_id = add_metadata(
        n=photon.batch_size, recursive=photon.recurse, detector=system.detector
    )
    add_system_data(simulation_id, system)

    # Prepare the list of parameter combinations
    param_combinations = list(itertools.product(*arrays.values()))
    keys = list(arrays.keys())

    # Prepare arguments for multiprocessing
    params = [
        (system, variable, photon, keys, values, output)
        for values in param_combinations
    ]

    # Process through all permutations of iterables
    num_workers = num_workers or mp.cpu_count()
    with mp.Pool(processes=num_workers) as pool:
        try:
            if pbar:
                results = list(
                    tqdm(
                        pool.imap(single_sim, params),
                        total=len(params),
                        desc=f"Sim ID: {simulation_id}",
                    )
                )
            else:
                results = list(pool.imap(single_sim, params))
        except Exception as e:
            logging.debug(f"Multiprocessing failed: {e}")
            results = []
        finally:
            pool.close()
            pool.join()

    for result in results:
        add_simulation_result(simulation_id, *result)

    c.close()
    return simulation_id


class LUT(BaseModel):
    simulation_id: int = latest_simulation_id
    dimensions: List[Dimensions] = [Dimensions.MU_S, Dimensions.MU_A, Dimensions.G]
    extrapolate: bool = False
    scale: float = 1

    _interpolator: Optional[RegularGridInterpolator] = PrivateAttr(default=None)

    class Config:
        ignored_types = (classOrInstanceMethod,)
        use_enum_values = True
        arbitrary_types_allowed = True

    def __call__(
        self, *values: Union[Real, np.ndarray], extrapolate: bool = None
    ) -> Union[Real, np.ndarray]:
        if not isinstance(values, tuple):
            values = (values,)
        if not len(values) <= len(self.dimensions):
            raise IndexError(
                f"LUT supports only up to {len(self.dimensions)}D. {self.dimensions}"
            )

        # Ensure all inputs are numpy arrays
        pts = [np.atleast_1d(v) for v in values]

        # Ensure all input arrays have the same shape for element-wise pairing
        input_shapes = [p.shape for p in pts]
        if len(set(input_shapes)) > 1:
            raise ValueError(
                f"Input arrays must have the same shape for element-wise pairing, got {input_shapes}"
            )

        # For unqueried dimensions, get full range from the database
        num_input_dims = len(values)
        for i in range(num_input_dims, len(self.dimensions)):
            c.execute(
                f"SELECT DISTINCT {self.dimensions[i]} FROM mclut WHERE simulation_id={self.simulation_id}"
            )
            pts.append(np.unique([row[0] for row in c.fetchall()]))

        # Pair elements
        query_pts = np.column_stack(pts)

        # Interpolation
        interpolator = self.interpolator
        interpolator.bounds_error = (
            not extrapolate if extrapolate is not None else not self.extrapolate
        )
        result = interpolator(query_pts)

        # Return result with the same shape as input arrays
        return result.reshape(input_shapes[0]) / self.scale

    @property
    def interpolator(self) -> RegularGridInterpolator:
        if self._interpolator is None:
            query = f"SELECT {', '.join(self.dimensions)}, output FROM mclut WHERE simulation_id={self.simulation_id}"
            c.execute(query)
            results = c.fetchall()
            if results is None or len(results) == 0:
                raise IOError(
                    f"No simulations found at id {self.simulation_id}. Run generate_lut "
                    f"and save results to lut.db before using lookup or try a "
                    f"different ID."
                )

            # If no exact match, check if parameters are within the bounds for interpolation
            *values, output = zip(*results)

            # Get data into a regular grid
            points = tuple(np.unique(val) for val in values)
            output = np.asarray(output).reshape(*[len(p) for p in points])

            # Interpolation/Extrapolate (if True)
            self._interpolator = RegularGridInterpolator(
                points,
                output,
                method="cubic",
                bounds_error=not self.extrapolate,
                fill_value=None,
            )

        return self._interpolator

    @classOrInstanceMethod
    def to_pandas(self, cls, simulation_id=None) -> pd.DataFrame:
        if simulation_id is None:
            simulation_id = latest_simulation_id
        query = f"SELECT mu_s, mu_a, g, output FROM mclut WHERE simulation_id = {simulation_id}"
        df = pd.read_sql_query(query, CON)
        return df

    @classOrInstanceMethod
    def list_available(self, cls, portion: str = "ALL") -> List[int]:
        c.execute(
            """
        SELECT DISTINCT id FROM mclut_simulations
        """
        )
        portion = ListPortion[portion.upper()].value
        ids = c.fetchall()
        available = []
        for id in ids:
            c.execute(
                f"""
            SELECT COUNT(*) from mclut WHERE simulation_id={id[0]}
            """
            )
            if c.fetchone()[0] > 1:
                available.append(id[0])
        return available[portion]

    @classOrInstanceMethod
    def get_layer_data(self, cls, simulation_id: Optional[int] = None) -> pd.DataFrame:
        if simulation_id is None:
            if hasattr(cls, "simulation_id"):
                simulation_id = cls.simulation_id
            else:
                simulation_id = latest_simulation_id
        query = f"SELECT * FROM fixed_layers WHERE simulation_id = {simulation_id}"
        df = pd.read_sql_query(query, CON)
        return df

    @classOrInstanceMethod
    def get_metadata(
        self, cls, simulation_id: Optional[int] = None, portion: str = "ALL"
    ) -> pd.DataFrame:
        query = f"SELECT * FROM mclut_simulations"
        portion = ListPortion[portion.upper()].value
        if simulation_id is None:
            if hasattr(cls, "simulation_id"):
                simulation_id = cls.simulation_id
                query += f" WHERE id = {simulation_id}"
        df = pd.read_sql_query(query, CON)
        return df[portion]

    def surface(self) -> Tuple[Real, Real, Real]:
        df = self.to_pandas(self.simulation_id)
        x = df[self.dimensions[0]].unique()
        y = df[self.dimensions[1]].unique()
        X, Y = np.meshgrid(x, y, indexing="ij")
        Z = np.reshape(df["output"], (len(x), len(y)))
        if hasattr(self, "scale"):
            Z /= self.scale
        return X, Y, Z



import random
import unittest

import numpy as np

from photon_canon import System, Medium, Illumination, Detector
from photon_canon.optics import Photon, IndexableProperty
from photon_canon.hardware import create_oblique_beams, create_cone_of_acceptance, ID, OD, THETA
from random import random


class TestPhoton(unittest.TestCase):
    def setUp(self) -> None:
        """
        Initializes a Photon instance with explicit settings for use in test cases.

        This method sets up a known Photon state, ensuring consistent
        initial conditions for testing initialization and behavior.
        """
        self.water = Medium(n=1.33, desc='water')
        self.tissue = Medium(n=1.53, mu_a=5, mu_s=100, g=0.85, desc='tissue')
        surroundings_n = 1.33

        sampler = create_oblique_beams((ID, OD), THETA)
        led = Illumination(pattern=sampler)
        detector = Detector(create_cone_of_acceptance(ID))

        self.system = System(
            self.water, 0.2,
            self.tissue, float('inf'),
            surrounding_n=surroundings_n,
            illuminator=led,
            detector=(detector, 0)
        )
        self.photon = Photon(
            wavelength=500,  # nm
            batch_size=100,
            system=self.system,
            directional_cosines=(0, 0, 1),  # Initially moving along +z
            location_coordinates=(0, 0, 0),
            weights=1.0,
            russian_roulette_constant=20,
            recurse=True,
            recursion_depth=0,
            recursion_limit=10,
            throw_recursion_error=True,
            keep_secondary_photons=True,

        )

    def test_initialization(self):
        """
        Verifies that all Photon attributes are correctly initialized.

        This test ensures that explicitly set attributes, as well as derived
        or automatically assigned attributes, match expected values.
        """

        # Explicit settings
        self.assertEqual(self.photon.wavelength, 500)
        self.assertEqual(self.photon.batch_size, 100)
        self.assertEqual(self.photon.system, self.system)
        self.assertEqual(self.photon.directional_cosines.shape, (100, 3))
        self.assertTrue(np.all(self.photon.directional_cosines == np.array([0, 0, 1])))
        self.assertIsInstance(self.photon.directional_cosines, IndexableProperty)
        self.assertEqual(self.photon.location_coordinates.shape, (100, 3))
        self.assertTrue(np.all(self.photon.location_coordinates == np.array([0, 0, 0])))
        self.assertEqual(self.photon.weights.shape, (100,))
        self.assertTrue(np.all(self.photon.weights == 1.0))
        self.assertEqual(self.photon.russian_roulette_constant, 20)
        self.assertTrue(self.photon.recurse)
        self.assertEqual(self.photon.recursion_depth, 0)
        self.assertEqual(self.photon.recursion_limit, 10)

        # Hidden settings
        self.assertEqual(self.photon.T, 0)
        self.assertEqual(self.photon.R, 0)
        self.assertEqual(self.photon.A, 0)
        self.assertEqual(self.photon.tir_count.shape, (100,))
        self.assertTrue(np.all(self.photon.tir_count == 0))
        self.assertEqual(self.photon.exit_location.shape, (100, 3))
        self.assertTrue(np.all(np.isnan(self.photon.exit_location)))
        self.assertEqual(self.photon.exit_direction.shape, (100, 3))
        self.assertTrue(np.all(np.isnan(self.photon.exit_direction)))
        self.assertEqual(self.photon.exit_weights.shape, (100,))
        self.assertTrue(np.all(np.isnan(self.photon.exit_weights)))
        self.assertEqual(self.photon.location_history.shape, (100, 3, 1))
        self.assertTrue(np.all(self.photon.location_history.squeeze() == self.photon.location_coordinates))
        self.assertEqual(self.photon.weights_history.shape, (100, 1))
        self.assertTrue(np.all(self.photon.weights_history.squeeze() == self.photon.weights))
        self.assertEqual(self.photon.cache_register.shape, (100,))
        self.assertFalse(np.any(self.photon.cache_register))
        self.assertEqual(self.photon._medium.shape, (100,))
        self.assertEqual(self.photon.at_interface.shape, (100,))
        self.assertEqual(len(self.photon.secondary_photons), 0)
        self.assertIsInstance(self.photon.secondary_photons, list)

    def test_directional_cosines(self) -> None:
        """
        Tests the filling and normalization of directional cosines.

        Ensures that the directional cosines setter correctly fills the batch:
        - When only one directional cosine is given
        and correctly normalizes:
        - The entire vector when assigned as a whole.
        - Individual components when updated via indexing.
        """
        # Test setter filling
        self.photon.directional_cosines = (1, 0, 0)
        self.assertEqual(self.photon.directional_cosines.shape, (100, 3))
        self.assertTrue(np.all(self.photon.directional_cosines == np.array([1, 0, 0])))

        # Test setter normalization
        self.photon.directional_cosines = (1, 1, 1)
        self.assertIsInstance(self.photon.directional_cosines, IndexableProperty)
        self.assertTrue(np.all(np.isclose(np.linalg.norm(self.photon.directional_cosines, axis=-1), 1)))
        self.assertTrue(np.all(self.photon.directional_cosines == 1 / np.sqrt(3)))

        # Test __setitem__ normalization (now normalizing by [1/sqrt(3), 1/sqrt(3), 1]
        self.photon.directional_cosines[:, 2] = 1
        self.assertIsInstance(self.photon.directional_cosines, IndexableProperty)
        self.assertTrue(np.all(np.isclose(np.linalg.norm(self.photon.directional_cosines, axis=-1), 1)))
        self.assertTrue(
            np.all(np.isclose(self.photon.directional_cosines, np.sqrt(3 * np.array([1 / 3, 1 / 3, 1]) / 5))))

    def test_location_coordinates(self):
        """
        Tests the setting and filling of location coordinates.

        Ensures that the coordinates setter correctly fills the batch when:
        - Only one coordinate set is given
        - The entire vector when assigned as a whole.
        - Individual components when updated via indexing.
        """
        # Test filling
        self.photon.location_coordinates = (0, 0, 0.1)
        self.assertEqual(self.photon.location_coordinates.shape, (100, 3))
        self.assertIsInstance(self.photon.location_coordinates, np.ndarray)
        self.assertIsInstance(self.photon.location_coordinates[0, 0], np.float64)
        self.assertTrue(np.all(self.photon.location_coordinates == np.array([0, 0, 0.1])))

        # Test batch setting
        self.location_coordinates = np.repeat(np.array([0, 0, 0])[np.newaxis, ...], 100, axis=0)
        self.assertEqual(self.photon.location_coordinates.shape, (100, 3))
        self.assertIsInstance(self.photon.location_coordinates, np.ndarray)
        self.assertIsInstance(self.photon.location_coordinates[0, 0], np.float64)
        self.assertTrue(np.all(self.location_coordinates == np.array([0, 0, 0])))

        # Test indexing
        self.photon.location_coordinates[:, 2] = 0.1
        self.assertIsInstance(self.photon.location_coordinates, np.ndarray)
        self.assertIsInstance(self.photon.location_coordinates[0, 0], np.float64)
        self.assertTrue(np.all(self.photon.location_coordinates == np.array([0, 0, 0.1])))

    def test_weight_and_russian_roulette(self):
        """
        Test the setting, filling, auto-killing, and roulette behavior of the weight setter.

        Ensures that the weight setter correctly fills the batch when:
        - Only one coordinate set is given
        - The entire vector when assigned as a whole.
        - Individual components when updated via indexing.
        Also ensures that the weight setter correctly puts photons below the threshold (0.005) through russian roulette
        with correct augmentation and with proper probability of augmentation.
        """
        # Weight can be reset in batch
        self.photon.weights = 0.5
        self.assertEqual(self.photon.weights.shape, (100,))
        self.assertTrue(np.all(self.photon.weights == 0.5))

        # Weight can be batched manually
        self.photon.weights = np.repeat(1.0, 100)
        self.assertEqual(self.photon.weights.shape, (100,))
        self.assertTrue(np.all(self.photon.weights == 1.0))

        # Weight cannot go negative
        self.photon.weights = -1
        self.assertTrue(np.all(self.photon.weights == 0))

        # 0 weight kills the photon
        self.assertTrue(self.photon.is_terminated)

        # Non-0 weight revives them
        self.photon.weights = 1
        self.assertFalse(self.photon.is_terminated)

        # Weight below threshold roulette's the photon
        self.photon.weights = 0.0001
        self.assertTrue(
            np.all(
                (self.photon.weights == 0.0001 * self.photon.russian_roulette_constant) | (self.photon.weights == 0)
            )
        )

        # Simulate russian roulette enough times that it should "hit" within epsilon of 1/constant.
        epsilon = 0.01
        n = 2 * int((1 / epsilon ** 2) / self.photon.batch_size)
        i = 0
        for _ in range(n):
            self.photon.weights = 0.0001
            i += np.sum(self.photon.weights != 0) / self.photon.batch_size
        self.assertTrue(np.isclose(i / n, 1 / self.photon.russian_roulette_constant, atol=epsilon))

    def test_absorb(self):
        """
        Tests the absorbing behavior. Should be 0 in water and 1 * albedo in tissue.
        """
        self.photon.absorb()
        self.assertTrue(np.all(self.photon.weights == 1))
        self.photon.location_coordinates = (0, 0, 0.3)  # Get into tissue
        self.photon.absorb()
        absorbed_weights = self.tissue.albedo_at(self.photon.wavelength)
        self.assertTrue(np.all(self.photon.weights == (1 - absorbed_weights)))
        self.assertTrue(np.isclose(self.photon.A, self.photon.batch_size * absorbed_weights))

    def test_move(self):
        """
        Tests the expected movement behavior with both auto moves to interfaces when scattering is 0 and movements
        inside scattering media.
        """
        # Simple move
        self.photon.move()  # Should move to first interface with headed_into = tissue
        self.assertTrue(np.all(self.photon.location_coordinates == np.array([0, 0, 0.2])))
        self.assertTrue(np.all(self.photon.headed_into() == self.tissue))
        self.assertTrue(np.all(self.photon.medium == self.tissue))
        self.assertTrue(np.all(self.photon.at_interface))

        # Check reflection at index mismatched interface
        spec_ref = np.abs((self.water.n - self.tissue.n) / (self.water.n + self.tissue.n)) ** 2
        new_weight = 1 - spec_ref
        self.assertTrue(np.all(np.isclose(self.photon.weights, new_weight)))
        self.assertIsNotNone(self.photon.secondary_photons)

        # Move across interfaces
        self.photon.move(0.1)
        self.assertTrue(np.all(np.isclose(self.photon.location_coordinates, np.array([0, 0, 0.3]))))
        self.assertTrue(np.all(self.photon.headed_into() == self.tissue))
        self.assertTrue(np.all(self.photon.medium == self.tissue))
        self.assertFalse(np.any(self.photon.at_interface))

        # Move at an angle
        dir_cos = np.array([1, 2, -2]) / 3  # Aimed towards interface
        self.photon.directional_cosines = dir_cos
        self.photon.move(float('inf'))  # move to interface
        self.assertFalse(np.any(np.isclose(self.photon.directional_cosines, dir_cos)))

        # Check refraction direction update
        n1_n2 = self.tissue.n / self.water.n
        new_dir = n1_n2 * dir_cos  # Scale in-plane components
        new_dir[2] = -np.cos(np.arcsin(n1_n2 * np.sin(np.arccos(-2 / 3))))  # Refract normal component
        self.assertTrue(np.all(np.isclose(self.photon.directional_cosines, new_dir)))

        # Check reflection at angled mismatched boundary
        spec_ref = np.abs(((self.tissue.n * dir_cos[2]) - (self.water.n * new_dir[2])) /
                          ((self.tissue.n * dir_cos[2]) + (self.water.n * new_dir[2]))) ** 2
        new_weight -= new_weight * spec_ref
        self.assertTrue(np.all(np.isclose(self.photon.weights, new_weight)))
        self.assertIsNotNone(self.photon.secondary_photons)

    def test_scatter(self):
        """
        Test that scatter returns new directions when in scattering media, and that it lets directions persis in
        non-scattering media and at interfaces.
        """
        # At non-scattering interface
        dir_cos = np.array([0, 0, 1])
        self.photon.scatter()
        self.assertTrue(np.all(self.photon.directional_cosines == dir_cos))

        # In non-scattering media
        self.photon.move(0.1)
        self.photon.scatter()
        self.assertTrue(np.all(self.photon.directional_cosines == dir_cos))

        # At interface with scatterer
        self.photon.move(0.1)
        self.photon.scatter()
        self.assertTrue(np.all(self.photon.directional_cosines == dir_cos))

        # In scattering media
        self.photon.move(0.1)
        self.photon.scatter()
        self.assertFalse(np.any(self.photon.directional_cosines == dir_cos))

class TestMedium(unittest.TestCase):
    def setUp(self):
        self.properties = {
            'desc': 'test',
            'display_color': 'gray',
            'n': random() + 1,  # Random index of refraction in (1, 2)
            'mu_s': 100 * random() + 40,  # Random reduced scatter coeff in (40, 140)
            'mu_a': 100 * random(),  # Random absorb coeff in (0, 100)
            'g': random(),  # Random anisotropy in (0, 1)
        }

        # Make medium test case
        self.medium = Medium(**self.properties)

    def test_init(self):
        for prop, val in self.properties.items():
            self.assertEqual(self.properties[prop], getattr(self.medium, prop))

    def test_mu_t(self):
        mu_t = self.properties['mu_s'] + self.properties['mu_a']
        self.assertEqual(self.medium.mu_t, mu_t)

    def test_albedo(self):
        albedo = self.properties['mu_a'] / (self.properties['mu_s'] + self.properties['mu_a'])
        self.assertEqual(self.medium.albedo, albedo)


class TestSystem(unittest.TestCase):
    def setUp(self):
        self.air = Medium(n=1.0, desc='air')
        self.tissue = Medium(n=1.4, desc='tissue')
        self.water = Medium(n=1.33, desc='water')
        self.system = System(self.air, 10, self.tissue, 20, self.water, 5, surrounding_n=1.0)
        self.sys = System()

    def test_system_initialization(self):
        interfaces = np.asarray([0, 10, 30, 35])
        self.assertEqual(len(self.system.layer), 5)  # Surroundings, Air, tissue, water, surroundings
        self.assertEqual(self.system.surroundings.n, 1.0)
        np.testing.assert_array_equal(self.system.boundaries[1:-1], interfaces)

    def test_in_medium(self):
        self.assertEqual(self.system.in_medium(-5), self.system.surroundings)  # Above the first layer
        self.assertEqual(self.system.in_medium(0), (self.system.surroundings, self.air))  # At first boundary
        self.assertEqual(self.system.in_medium(5), self.air)  # Within the first layer
        self.assertEqual(self.system.in_medium(10), (self.air, self.tissue))  # At second interfaces
        self.assertEqual(self.system.in_medium(15), self.tissue)  # Within the second layer
        self.assertEqual(self.system.in_medium(35), (self.water, self.system.surroundings))  # At last boundary
        self.assertEqual(self.system.in_medium(40), self.system.surroundings)  # Below last layer

    def test_interface_crossed(self):
        pass
        # Crossing one interfaces cleanly, should return that interfaces
        # zs = [5, 25]
        #
        # interface, plane = self.system.interface_crossed(*zs)
        # self.assertEqual(interface, (self.air, self.tissue))
        # self.assertEqual(plane, 10)  # First interfaces at z = 10

        # Crossing two interfaces in positive direction, should return shallowest
        # interface, plane = self.system.interface_crossed(5, 25)  # Crossing from air to tissue (interfaces at 10 and )

        # Crossing two interfaces in negative direciton, should return deepest

        # Crossing no interfaces, should return

        # Crossing out of the media

        # Start at interfaces and don't cross

        # Start at interfaces and cross another

    class TestPhoton(unittest.TestCase):
        def setUp(self):
            self.tissue = Medium(n=1.4, mu_s=2, mu_a=0.5, g=0.8, desc='tissue')
            self.water = Medium(n=1.33, mu_s=1.5, mu_a=0.3, g=0.7, desc='water')
            self.system = System(self.tissue, 20, self.water, 30, surrounding_n=1.0)
            self.photon = Photon(wavelength=500, system=self.system, location_coordinates=(0, 0, 10))

        def test_photon_initialization(self):
            self.assertEqual(self.photon.wavelength, 500)
            self.assertEqual(tuple(self.photon.location_coordinates), (0, 0, 10))
            self.assertEqual(tuple(self.photon.directional_cosines), (0, 0, 1))

        def test_photon_weight_behavior(self):
            self.photon.weights = 0.004  # Triggers Russian roulette
            self.assertIn(self.photon.weights, [0, 0.004 * self.photon.russian_roulette_constant])

        def test_photon_medium(self):
            self.assertEqual(self.photon.medium, self.tissue)
            self.photon.location_coordinates = np.array([0, 0, 25])
            self.assertEqual(self.photon.medium, self.water)

        def test_photon_movement(self):
            initial_position = self.photon.location_coordinates.copy()
            self.photon.move()
            self.assertFalse(np.array_equal(initial_position, self.photon.location_coordinates))

        def test_photon_absorption(self):
            initial_weights = self.photon.weights
            self.photon.absorb()
            self.assertLess(self.photon.weights, initial_weights)
            self.assertGreater(self.photon.A, 0)

        def test_photon_scattering(self):
            initial_direction = self.photon.directional_cosines.copy()
            self.photon.scatter()
            self.assertFalse(np.array_equal(initial_direction, self.photon.directional_cosines))


if __name__ == '__main__':
    unittest.main()



import warnings
from numbers import Real
from pathlib import Path
from typing import Tuple, Iterable, Union, Callable

import sqlite3

from .import_utils import np

# Setup default database
db_dir = Path.home() / ".photon_canon"
db_dir.mkdir(parents=True, exist_ok=True)
db_path = db_dir / "lut.db"
CON = sqlite3.connect(db_path)
c = CON.cursor()

class OpticalPropertyError(Exception):
    """
    Custom exception raised when invalid or undefined optical properties are encountered.

    This can be used to signal errors related to scattering or absorption values, LUT mismatches,
    or other property-related modeling issues.
    """
    pass

# This block attempts to retrieve the latest simulation ID from the mclut_simulations table.
# If the table does not exist or the query fails, it sets latest_simulation_id to None and emits a warning.
try:
    c.execute("SELECT max(id) FROM mclut_simulations")
    latest_simulation_id = c.fetchone()[0]
except sqlite3.OperationalError:
    latest_simulation_id = None
    warnings.warn("No default LUT found. Simulate one if you have not.")


def simulate(system: "System", n: int, **kwargs) -> Tuple[float, float, float]:
    """
    Runs a Monte Carlo simulation using the provided system configuration and number of photons.

    The simulation is executed on a `System` object, which models beam propagation and optical interactions.
    Results include the transmittance (T), reflectance (R), and absorption (A) coefficients.

    :param system: The optical system object capable of simulating photon transport.
    :type system: System
    :param n: Number of photons to simulate.
    :type n: int
    :param kwargs: Additional keyword arguments passed to `system.beam()`.
    :type kwargs: Any

    :return: Tuple containing transmittance (T), reflectance (R), and absorption (A).
    :rtype: Tuple[float, float, float]
    """
    photons = system.beam(n=n, **kwargs)
    photons.simulate()
    return photons.T, photons.R, photons.A


def sample_spectrum(wavelengths: Iterable[Real], spectrum: Iterable[Real]):
    """
    Samples a wavelength from a spectral probability distribution.

    The function normalizes the input spectrum to form a probability density function (PDF),
    computes the cumulative distribution function (CDF), and uses inverse transform sampling
    to randomly select a wavelength based on the distribution.

    :param wavelengths: Array-like list of wavelengths.
    :type wavelengths: Iterable[Real]
    :param spectrum: Array-like list of spectral intensities or probabilities corresponding to the wavelengths.
    :type spectrum: Iterable[Real]

    :return: A single sampled wavelength based on the input spectrum.
    :rtype: float
    """
    wavelengths = np.asarray(wavelengths)
    spectrum = np.asarray(spectrum)

    # Normalize PDF
    spectrum /= np.sum(spectrum)

    # Compute CDF
    cdf = np.cumsum(spectrum)

    # Take random sample
    i = np.random.uniform(0, 1)

    # Interpolate value of sample from CDF
    return np.interp(i, cdf, wavelengths)


def model_reflectance(
    model: Callable[[float, float, ...], Union[float, np.ndarray]],
    mu_s: np.ndarray[float],
    mu_a: np.ndarray[float],
    vectorize: bool = False,
    **kwargs
) -> np.ndarray[float]:
    """
    Generalizes reflectance modeling by applying a provided model function to arrays of scattering and absorption coefficients.

    This function computes modeled reflectance based on the given `model`, scattering coefficients (`mu_s`),
    and absorption coefficients (`mu_a`). If `vectorize` is set to True, the function assumes that the model
    supports vectorized input and calls it once for all values. Otherwise, it loops through each pair
    of (`mu_s`, `mu_a`) values and computes the reflectance individually.

    :param model: A function that accepts `mu_s`, `mu_a`, and optional keyword arguments to return reflectance.
                  Must return either a float or a NumPy array.
    :type model: Callable[[float, float, ...], Union[float, np.ndarray]]
    :param mu_s: An array of scattering coefficients.
    :type mu_s: np.ndarray[float]
    :param mu_a: An array of absorption coefficients.
    :type mu_a: np.ndarray[float]
    :param vectorize: Whether to evaluate the model in a vectorized fashion. If False, uses a loop.
    :type vectorize: bool
    :param kwargs: Additional keyword arguments to pass to the model function.
    :type kwargs: Any

    :return: Array of modeled reflectance values corresponding to each (`mu_s`, `mu_a`) pair.
    :rtype: np.ndarray[float]
    """
    if vectorize:
        rd = model(mu_s, mu_a, **kwargs)
    else:
        rd = []
        for mus, mua in zip(mu_s, mu_a):
            rd.append(model(mus, mua, **kwargs))
        rd = np.array(rd)
    return rd.squeeze()



import importlib.resources
from numbers import Real
from typing import Union, Iterable, Tuple, Callable

import pandas as pd

from ..import_utils import np, NDArray, interp1d
from ..utils import model_reflectance, OpticalPropertyError

# Read data file for function usage
with importlib.resources.open_text("photon_canon.data", "hbo2_hb.tsv") as f:
    df = pd.read_csv(f, sep="\t", skiprows=1)[1:]
wl, hbo2, dhb = df["lambda"], df["hbo2"], df["hb"]

# Extract arrays and convert to mu_a units (cm-1)
wl = np.array([float(w) for w in wl])
hbo2 = np.array([float(h) for h in hbo2])
dhb = np.array([float(h) for h in dhb])
eps = np.log(10) * np.stack((hbo2, dhb)) / 64500

# Calculate some defaults (good for initial guesses)
tHb = 1
sO2 = 0.5

class ConcentrationError(OpticalPropertyError):
    """Special error for concentration validation b/c it comes up all the time in Jacobian calculations for fitting"""
    pass

def validate_concentrations_and_spectra(ci, epsilons, wavelength):
    # Helper function to check if concentrations are valid (non-negative)
    def check_concentrations_valid(ci):
        if isinstance(ci, (list, tuple, np.ndarray)):
            if np.any(np.array(ci) < 0):
                raise ConcentrationError("Concentrations cannot be negative")
        elif isinstance(ci, (int, float)):
            if ci < 0:
                raise ConcentrationError("Concentrations cannot be negative")

    # Check if lengths match for list-like concentrations and epsilons
    def check_lengths_match(ci, epsilons):
        if isinstance(ci, (list, tuple, np.ndarray)):
            if len(ci) != len(epsilons):
                raise OpticalPropertyError(f"Length mismatch: {len(ci)} concentrations vs {len(epsilons)} epsilons")
        elif isinstance(ci, (int, float)):
            if isinstance(epsilons[0], (list, tuple, np.ndarray)) and len(epsilons) != 1:
                raise OpticalPropertyError(f"Expected 1 epsilon, but got {len(epsilons)}")

    # Check if wavelengths and epsilons match up
    def check_wavelength_epsilon_match(wavelength, epsilons):
        msg = f"A spectrum of molar absorptivity must be included with each spectrum. You gave {len(wavelength)} wavelengths but molar absorptivity had {len(epsilons[0])} elements."
        if isinstance(epsilons[0], (list, tuple, np.ndarray)):
            if not all(len(e) == len(wavelength) for e in epsilons):
                raise OpticalPropertyError(msg)
        elif isinstance(epsilons[0], (int, float)) and len(epsilons) != len(wavelength):
            raise OpticalPropertyError(msg)

        check_lengths_match(ci, epsilons)

        check_wavelength_epsilon_match(wavelength, epsilons)

        check_concentrations_valid(ci)

def calculate_mus(
    a: Real = 1,
    b: Real = 1,
    ci: Union[Real, Iterable[Real]] = (tHb * sO2, tHb * (1 - sO2)),
    epsilons: Union[Iterable[Real], Iterable[Iterable[Real]]] = eps,
    wavelength: Union[Real, Iterable[Real]] = wl,
    wavelength0: Real = 650,
    force_feasible: bool = True,
) -> Union[Tuple[Real, Real, Real], Tuple[NDArray, NDArray, NDArray]]:
    # Check cs and epsilons match up and are feasible
    try:
        validate_concentrations_and_spectra(ci, epsilons, wavelength)
    except ConcentrationError as e:
        if force_feasible:
            # Stash infeasible results into error
            raise e(calculate_mus(a, b, ci, epsilons, wavelength, wavelength0, force_feasible=False))

    # Array everything as needed
    wavelength = np.asarray(wavelength)  # Wavelengths of measurements (nm)
    mu_s = a * (wavelength / wavelength0) ** -b  # Reduced scattering coefficient, cm^-1

    # Unpack list of spectra (if it is a list)
    if isinstance(epsilons[0], (tuple, list, np.ndarray)):
        epsilons = np.asarray(
            [np.asarray(spectrum) for spectrum in epsilons]
        )  # Molar absorptivity (L/(mol cm))
    else:
        epsilons = np.asarray(epsilons)

    # Reshape concentrations (if multiple)
    if isinstance(ci, (list, tuple, np.ndarray)):
        ci = np.asarray(ci)
        ci = ci.reshape(-1, 1)
    mu_a = np.sum(ci * epsilons, axis=0)  # Absorption coefficient, cm^-1
    return mu_s, mu_a, wl


def hemoglobin_mus(
    a: Real = 1,
    b: Real = 1,
    t: Real = tHb,
    s: Real = sO2,
    wavelengths: Iterable[Real] = wl,
    force_feasible: bool = True,
) -> Union[Tuple[Real, Real, Real], Tuple[NDArray, NDArray, NDArray]]:
    """
    Computes the reduced scattering coefficient (μs') for hemoglobin solutions
    based on given absorption coefficients of oxyhemoglobin (HbO2) and deoxyhemoglobin (Hb).

    This function interpolates the extinction coefficients of HbO2 and Hb
    at specified wavelengths using cubic interpolation and calculates the
    corresponding μs' values.

    :param force_feasible: Option to prevent impossible values (like negative concentrations) (default: True)
    :type force_feasible: bool
    :param a: Scaling factor for the reduced scattering coefficient (default: 1).
    :type a: Real
    :param b: Scattering power exponent (default: 1).
    :type b: Real
    :param t: Total hemoglobin concentration (tHb) in g/L (default: `tHb`).
    :type t: Real
    :param s: Oxygen saturation (sO2) as a fraction (0 to 1) (default: `sO2`).
    :type s: Real
    :param wavelengths: Array of wavelengths (in nm) at which to compute the values (default: `wl`).
    :type wavelengths: Iterable[Real]

    :return: A tuple containing:
             - The reduced scattering coefficient (μs') at each wavelength.
             - The interpolated extinction coefficient for THb.
             - The wavelengths of the measures (passed unchanged from input, useful for defaulting and reusing)
    :rtype: Tuple[Real, Real, Real] or Tuple[NDArray, NDArray, NDArray]

    :notes:
        - The extinction coefficients for HbO2 and Hb are interpolated using cubic
          interpolation from a predefined dataset.
        - Extrapolation is used for wavelengths outside the dataset range.
        - The calculation assumes a power-law dependence on wavelength for scattering.
    """
    hbo2_interp = interp1d(wl, eps[0], kind="cubic", fill_value="extrapolate")
    dhb_interp = interp1d(wl, eps[1], kind="cubic", fill_value="extrapolate")
    epsilons = (hbo2_interp(wavelengths), dhb_interp(wavelengths))
    ci = (s * t, (1 - s) * t)
    return calculate_mus(
        a, b, ci, epsilons, wavelengths, wavelength0=650, force_feasible=force_feasible
    )


def model_from_hemoglobin(
    model: Callable[[float, float, ...], float],
    wavelengths: np.ndarray[Union[int, float]],
    a: np.ndarray[float],
    b: np.ndarray[float],
    t: np.ndarray[float],
    s: np.ndarray[float],
    **kwargs
) -> np.ndarray[float]:
    """Forwarding function to streamline Hb/sO2 modelling. This function takes biological parameters and turns them
    optical properties, then forwards those to model reflectance and returns that reflectance. In short, biological
    properties get converted to reflectance using the given model."""
    mu_s, mu_a, _ = hemoglobin_mus(a, b, t, s, wavelengths)
    return model_reflectance(model, mu_s, mu_a, **kwargs)



import functools
from typing import Tuple, Callable

from scipy.integrate import quad

from ..import_utils import np

# TODO: Check that everything in these functions will vectorize


def fresnel_equation(alpha_i: float = 0, n_rel: float = 1) -> float:
    cos_alpha_t = np.sqrt(1 - np.sin(alpha_i / n_rel) ** 2)
    rs = (
        np.abs(
            (np.cos(alpha_i) - n_rel * cos_alpha_t)
            / (np.cos(alpha_i) + n_rel * cos_alpha_t)
        )
        ** 2
    )
    rp = (
        np.abs(
            (cos_alpha_t - n_rel * np.cos(alpha_i))
            / (cos_alpha_t + n_rel * np.cos(alpha_i))
        )
        ** 2
    )
    specular_reflection = 0.5 * (rs + rp)
    return specular_reflection


def cylindrical_distance(
    rtz: Tuple[float, float, float], rtz_prime: Tuple[float, float, float]
) -> float:
    r, theta, z = rtz
    r_prime, theta_prime, z_prime = rtz_prime
    return np.sqrt(
        (r - r_prime) ** 2
        + r * r_prime * np.cos(theta - theta_prime)
        + (z - z_prime) ** 2
    )


def diffusion_approximation(
    mu_s: float,
    mu_a: float,
    r: float = 1,
    alpha_i: float = 0,
    n_rel: float = 1.33,
    g: float = 0.9,
) -> float:
    # Reduced scattering coefficient
    mu_s_prime = (1 - g) * mu_s

    # Transport coefficient
    mu_t_prime = mu_s_prime + mu_a

    # Mean free path length
    l_prime = 1 / mu_t_prime

    # Transport albedo
    a_prime = mu_s_prime / mu_t_prime

    # Azimuthal angle of source and image
    theta, theta_prime = 0, 0

    # Normal incidence
    if alpha_i == 0:
        # Diffusion coefficient
        D = 1 / (3 * (mu_s_prime + mu_a))

        # Effective attenuation coefficient
        mu_eff = np.sqrt(mu_a / D)

        # Effective source distance
        z_prime = l_prime

        # Center of diffuse reflection
        r_prime = 0

    # Oblique incidence corrections
    else:
        # Diffusion coefficient
        D = 1 / (3 * (mu_s_prime + 0.35 * mu_a))

        # Effective attenuation coefficient
        mu_eff = np.sqrt(mu_a / D)

        # Transmission angle (Snell's law)
        alpha_t = np.arcsin(np.sin(alpha_i) / n_rel)

        # Center of far diffuse reflectance
        r_prime = 3 * D * np.sin(alpha_t)

        # The new effective source distance as a result of the oblique incidence
        z_prime = r_prime / np.tan(alpha_t)

    # Fresnel reflection
    Rsp = fresnel_equation(alpha_i, n_rel)

    # Extrapolated boundary location
    z_b = -2 * D

    # Distance between observation point (r, 0, 0) and original source (0, 0, z_prime)
    rho1 = cylindrical_distance((r, theta, 0), (r_prime, theta_prime, z_prime))

    # Distance between observation point (r, 0, 0) and image source point (0, 0, -z_prime - 2*z_b)
    rho2 = cylindrical_distance(
        (r, theta, 0), (r_prime, theta_prime, -z_prime - 2 * z_b)
    )

    # Diffuse reflectance at r
    Rd = (
        a_prime
        * z_prime
        * (1 + mu_eff * rho1)
        * np.exp(-mu_eff * rho1)
        / (4 * np.pi * rho1**3)
    ) + (
        a_prime
        * (z_prime + 4 * D)
        * (1 + mu_eff * rho2)
        * np.exp(-mu_eff * rho2)
        / (4 * np.pi * rho2**3)
    )
    Rd *= 1 - Rsp
    return Rd


def create_diffusion_approximation(
    r: float = 1, alpha_i: float = 0, n_rel: float = 1.33, g: float = 0.9
) -> Callable[[float, float], float]:
    return functools.partial(
        diffusion_approximation, r=r, alpha_i=alpha_i, n_rel=n_rel, g=g
    )


def create_integrated_diffusion_approximation(
    r_range: Tuple[float, float],
    alpha_i: float = 0,
    n_rel: float = 1.33,
    g: float = 0.9,
) -> Callable[[float, float], float]:
    def integrand(r: float, mu_s: float, mu_a: float) -> float:
        return (
            2
            * np.pi
            * r
            * diffusion_approximation(
                mu_s=mu_s, mu_a=mu_a, r=r, alpha_i=alpha_i, n_rel=n_rel, g=g
            )
        )

    def integrated_diffusion_approximation(mu_s: float, mu_a: float) -> float:
        result, err = quad(integrand, r_range[0], r_range[1], args=(mu_s, mu_a))
        return result

    return integrated_diffusion_approximation
